/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

/* The following description is used by demos/make-demo-list.pl
 * to compile a list of all tests. */
/* DESCRIPTION
 * By-Id (backref) test
 *
 * Function: demo_by_id1
 */


#include <gtk/gtk.h>
#include <gdam/gdambuffer.h>
#include "../gdamchannel.h"

typedef struct _DemoInfo {
	GdamChannel* channel;
	GtkWindow* top_level;
	GtkLabel* filename;
	GtkFileSelection* file_selection;
	guint mixer_id;
	guint source_id;
	gboolean making_request;
} DemoInfo;

static void chose_file(GtkButton* button, DemoInfo* info)
{
	char* filename;
	g_return_if_fail(info->file_selection != NULL);
	filename = gtk_file_selection_get_filename(info->file_selection);
	gtk_label_set_text(info->filename, filename);
}

static void filesel_button_clicked(GtkButton* button, DemoInfo* info)
{
	if (info->file_selection) {
		GtkObject* to_kill = GTK_OBJECT(info->file_selection);
		info->file_selection = NULL;
		gtk_object_destroy(to_kill);
	} else {
		GtkWidget* filesel = gtk_file_selection_new("pick your mp3");
		GtkFileSelection* file_select = (GtkFileSelection*)filesel;
		info->file_selection = file_select;
		gtk_signal_connect(GTK_OBJECT(file_select->ok_button),
		                   "clicked",
		                   (GtkSignalFunc) chose_file,
				   info);
		/*gtk_signal_connect_object(
		                   GTK_OBJECT(file_select->cancel_button),
		                   "clicked",
				   (GtkSignalFunc) gtk_widget_destroy,
				   GTK_OBJECT(file_select));*/
		gtk_widget_show(GTK_WIDGET(file_select));
	}
}

static void play_callback(GtkButton* button, DemoInfo* info)
{
	gchar* filename;
	if (info->source_id != 0 || info->making_request)
		return;

	info->making_request = TRUE;
	gtk_label_get(info->filename, &filename);
	gdam_debug("create gain %s", filename);
	info->source_id = gdam_channel_create_source_generic(
			info->channel,
			"GdamSeekableMp3",
			1,
			"filename",
			GDAM_ARG_TYPE_STRING,
			filename);
	g_assert(info->source_id != 0);
	gdam_debug("gdam_channel_mix_add being called...");
	if (!gdam_channel_mix_add(info->channel,
	                          info->mixer_id,
				  info->source_id,
				  -1))
		g_warning("gain_add failed");
	info->making_request = FALSE;
}

static void inject_callback(GtkButton* button, DemoInfo* info)
{
	g_return_if_fail(info->source_id != 0);
	info->source_id = gdam_channel_create_source_generic(info->channel,
			      "GdamGain",
			      2,
			      "gain",
			      GDAM_ARG_TYPE_DOUBLE,
			      0.5,
			      "filtered",
			      GDAM_ARG_TYPE_INT,
			      info->source_id);
	if (!gdam_channel_mix_add(info->channel,
	                          info->mixer_id,
				  info->source_id,
				  -1))
		g_warning("gain_add failed");
}
static void inject_package_callback(GtkButton* button, DemoInfo* info)
{
	GdamRequest** requests = g_new(GdamRequest*, 2);
	gboolean mix_add_return;
	GdamRequestGetIntStruct create_source_return, package_return;

	g_return_if_fail(info->source_id != 0);

	requests[0] = gdam_request_create_source(
			"GdamGain",
			gdam_request_get_int,
			gdam_request_get_failure,
			&create_source_return,
			2,
			"gain",
			GDAM_ARG_TYPE_DOUBLE,
			0.5,
			"filtered",
			GDAM_ARG_TYPE_INT,
			info->source_id
		      );
	requests[1] = gdam_request_add_mix(
			info->mixer_id,
			-1,		/* subsource id */
			-1,		/* when */
			gdam_request_get_success,
			gdam_request_get_failure,
			&mix_add_return
		      );
	if (!gdam_channel_process_request(info->channel,
	                                  gdam_request_package_v(
						2,
						requests,
						gdam_request_get_int,
						gdam_request_get_failure,
						&package_return))) {
		g_warning("error issuing package request");
		return;
	}
	g_return_if_fail(mix_add_return);
	g_return_if_fail(create_source_return.was_success);
	g_return_if_fail(package_return.was_success);
	info->source_id = create_source_return.rv;
}


static void stop_callback(GtkButton* button, DemoInfo* info)
{
	if (info->source_id == 0 || info->making_request)
		return;
	if (!gdam_channel_detach_source(info->channel, info->source_id))
		g_warning("detach source failed");
	info->source_id = 0;
}


/* Helper + macro to cut down repetition in demo1_init(). */
static GtkWidget* _generic_button(const char* label, GtkSignalFunc callback,
                                 gpointer user_data)
{
	GtkWidget* button = gtk_button_new_with_label(label);
	gtk_signal_connect(GTK_OBJECT(button), "clicked", callback, user_data);
	gtk_widget_show(GTK_WIDGET(button));
	return button;
}

static void gain_changed(GtkAdjustment* adjustment, DemoInfo* info)
{
	if (info->source_id == 0)
		return;
	if (!gdam_channel_source_set_arg(info->channel,
	                                 info->source_id, 
					 "rate",
					 GDAM_ARG_TYPE_DOUBLE,
                                         adjustment->value))
		return;
}

#define generic_button(label, callback)				\
		_generic_button(label, GTK_SIGNAL_FUNC(callback), info)

void demo_by_id1(GdamChannel*           channel,
                guint                  mixer_id)
{
	DemoInfo* info = g_new(DemoInfo, 1);
	GtkWidget* hbox;
	GtkWidget* vbox;
	GtkAdjustment* adjustment;
	GtkWidget* scroll;

	vbox = gtk_vbox_new(FALSE, 0);
	hbox = gtk_hbox_new(FALSE, 0);

	info->channel = channel;
	info->top_level = (GtkWindow*) gtk_window_new(GTK_WINDOW_TOPLEVEL);
	info->filename = (GtkLabel*) gtk_label_new("???");
	info->file_selection = NULL;
	info->making_request = FALSE;
	info->mixer_id = mixer_id;
	info->source_id = 0;

	gtk_widget_show(GTK_WIDGET(info->filename));
	gtk_box_pack_start_defaults(GTK_BOX(hbox), GTK_WIDGET(info->filename));
	gtk_box_pack_start_defaults(GTK_BOX(hbox),
	                  generic_button("File...", filesel_button_clicked));
	gtk_widget_show(hbox);
	gtk_box_pack_start_defaults(GTK_BOX(vbox), GTK_WIDGET(hbox));


	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start_defaults(GTK_BOX(hbox),
	                        generic_button("Play", play_callback));
	gtk_box_pack_start_defaults(GTK_BOX(hbox),
	                        generic_button("Inject", inject_callback));
	gtk_box_pack_start_defaults(GTK_BOX(hbox),
		generic_button("InjectAtomic", inject_package_callback));
	gtk_box_pack_start_defaults(GTK_BOX(hbox),
	                        generic_button("Stop", stop_callback));
	gtk_widget_show(hbox);
	gtk_box_pack_start_defaults(GTK_BOX(vbox), GTK_WIDGET(hbox));


	adjustment = GTK_ADJUSTMENT(gtk_adjustment_new(
					1.0, 0.0, 3.0,	/* default, min, max */
					0.01, 0.1, 1.0));
	gtk_signal_connect(GTK_OBJECT(adjustment),
	                   "value_changed",
			   GTK_SIGNAL_FUNC(gain_changed),
			   info);
	scroll = gtk_hscrollbar_new(adjustment);
	gtk_range_set_update_policy(GTK_RANGE(scroll), GTK_UPDATE_CONTINUOUS);
	gtk_widget_show(scroll);
	gtk_box_pack_start_defaults(GTK_BOX(vbox), GTK_WIDGET(scroll));

	gtk_widget_show(vbox);
	gtk_container_add(GTK_CONTAINER(info->top_level), GTK_WIDGET(vbox));
	gtk_widget_show(GTK_WIDGET(info->top_level));
}
