/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
/* The following description is used by demos/make-demo-list.pl
 * to compile a list of all tests. */
/* DESCRIPTION
 * A test of caching.
 *
 * Function: demo_cache0
 */


#include <gtk/gtk.h>
#include <gdam/gdambuffer.h>
#include "../gdamchannel.h"

typedef struct _DemoInfo {
	GdamChannel* channel;

	gboolean caching;
	gboolean making_mp3_request;

	guint cache_id;

	guint mixer_id;

	GtkWindow* top_level;
	GtkLabel* filename;
	GtkLabel* state;
	GtkFileSelection* file_selection;
	GtkToggleButton* loop_toggle;

	GList* source_infos;
	GtkWidget* button_box;
} DemoInfo;


typedef struct _SourceInfo {
	guint source_id;
	guint ended_id;
	GtkButton* button;
	DemoInfo* info;
} SourceInfo;


static void chose_file(GtkButton* button, DemoInfo* info)
{
	char* filename;
	g_return_if_fail(info->file_selection != NULL);
	filename = gtk_file_selection_get_filename(info->file_selection);
	gtk_label_set_text(info->filename, filename);
}

static void filesel_button_clicked(GtkButton* button, DemoInfo* info)
{
	if (info->file_selection) {
		GtkObject* to_kill = GTK_OBJECT(info->file_selection);
		info->file_selection = NULL;
		gtk_object_destroy(to_kill);
	} else {
		GtkWidget* filesel = gtk_file_selection_new("pick your mp3");
		GtkFileSelection* file_select = (GtkFileSelection*)filesel;
		info->file_selection = file_select;
		gtk_signal_connect(GTK_OBJECT(file_select->ok_button),
		                   "clicked",
		                   (GtkSignalFunc) chose_file,
				   info);
		/*gtk_signal_connect_object(
		                   GTK_OBJECT(file_select->cancel_button),
		                   "clicked",
				   (GtkSignalFunc) gtk_widget_destroy,
				   GTK_OBJECT(file_select));*/
		gtk_widget_show(GTK_WIDGET(file_select));
	}
}

static void kill_source_info(SourceInfo* info)
{
	DemoInfo* demo = info->info;
	if (info->source_id) {
		guint source_id = info->source_id;
		info->source_id = 0;
		gdam_channel_detach_source(demo->channel, source_id);
	}
	if (info->ended_id) {
		int id = info->ended_id;
		info->ended_id = 0;
		gdam_channel_unsubscribe(demo->channel, id);
	}
	if (info->button) {
		GtkWidget* widget = GTK_WIDGET(info->button);
		info->button = NULL;
		gtk_widget_destroy(widget);
	}
}

static void ended_callback(gpointer user_data)
{
	SourceInfo* info = (SourceInfo*)user_data;
	if (info->source_id != 0) {
		info->ended_id = 0;
		info->source_id = 0;
		kill_source_info(info);
	}
}

static void stop_clicked_callback(gpointer user_data)
{
	SourceInfo* info = (SourceInfo*)user_data;
	if (info->source_id != 0)
		kill_source_info(info);
}

static void done_caching(gpointer i)
{
	DemoInfo* info = (DemoInfo*)i;
	g_message("done caching");
	info->caching = FALSE;
}

static void start_caching_callback(GtkObject* unused,DemoInfo* info)
{
	char* filename;
	guint mp3_id;
	if (info->cache_id != 0) {
		guint cache_id = info->cache_id;
		info->cache_id = 0;
		gdam_channel_uncache(info->channel, cache_id);
	}
	if (info->caching)
		return;
	info->caching = TRUE;
	gtk_label_get(info->filename, &filename);
	mp3_id = gdam_channel_create_source_generic(
			info->channel,
			"GdamMp3",
			1,
			"filename",
			GDAM_ARG_TYPE_STRING,
			filename);
	info->cache_id = gdam_channel_start_caching(info->channel,
						    mp3_id,
				                    done_caching,
				                    info);
}

static void play_callback(GtkButton* button, DemoInfo* info)
{
	gchar* filename;
	guint id;
	SourceInfo* source_info;
	if (info->caching || !info->cache_id)
		return;

	info->making_mp3_request = TRUE;
	gtk_label_get(info->filename, &filename);
	id = gdam_channel_create_source_generic(
				info->channel,
				"GdamCache",
				2,
				"buffer_id",
				GDAM_ARG_TYPE_INT,
				info->cache_id, 
				"loop",
				GDAM_ARG_TYPE_INT,
				info->loop_toggle->active 
				);
	g_assert(id != 0);

	source_info = g_new(SourceInfo, 1);
	source_info->source_id = id;
	source_info->info = info;
	if (!gdam_channel_mix_add(info->channel,
	                          info->mixer_id,
				  id,
				  -1))
		g_warning("mix_add failed");
	source_info->ended_id = gdam_channel_subscribe_source_ended(
						info->channel, 
						id,
						ended_callback,
						source_info);
	source_info->button = (GtkButton*)gtk_button_new_with_label(
					g_strdup_printf("source %u", id));
	gtk_box_pack_start_defaults(GTK_BOX(info->button_box),
	                            GTK_WIDGET(source_info->button));
	gtk_widget_show(GTK_WIDGET(source_info->button));

	gtk_signal_connect_object(GTK_OBJECT(source_info->button),
	                          "clicked",
			          GTK_SIGNAL_FUNC(stop_clicked_callback),
				  (GtkObject*) source_info);


	info->making_mp3_request = FALSE;
}


/* Helper + macro to cut down repetition in demo1_init(). */
static GtkWidget* _generic_button(const char* label, GtkSignalFunc callback,
                                 gpointer user_data)
{
	GtkWidget* button = gtk_button_new_with_label(label);
	gtk_signal_connect(GTK_OBJECT(button), "clicked", callback, user_data);
	gtk_widget_show(GTK_WIDGET(button));
	return button;
}
#define generic_button(label, callback)				\
		_generic_button(label, GTK_SIGNAL_FUNC(callback), info)

void demo_cache0(GdamChannel*           channel,
                guint                  mixer_id)
{
	DemoInfo* info = g_new(DemoInfo, 1);
	GtkWidget* hbox;
	GtkWidget* vbox;
	vbox = gtk_vbox_new(FALSE, 0);
	hbox = gtk_hbox_new(FALSE, 0);

	info->caching = FALSE;
	info->cache_id = 0;
	info->channel = channel;
	info->top_level = (GtkWindow*) gtk_window_new(GTK_WINDOW_TOPLEVEL);
	info->filename = (GtkLabel*) gtk_label_new("???");
	info->state = (GtkLabel*) gtk_label_new("not started");
	info->file_selection = NULL;
	info->making_mp3_request = FALSE;
	info->mixer_id = mixer_id;

	gtk_widget_show(GTK_WIDGET(info->filename));
	gtk_box_pack_start_defaults(GTK_BOX(hbox), GTK_WIDGET(info->filename));
	gtk_box_pack_start_defaults(GTK_BOX(hbox),
	                  generic_button("File...", filesel_button_clicked));
	gtk_widget_show(hbox);
	gtk_box_pack_start_defaults(GTK_BOX(hbox), GTK_WIDGET(info->state));
	gtk_widget_show(GTK_WIDGET(info->state));

	gtk_box_pack_start_defaults(GTK_BOX(vbox), GTK_WIDGET(hbox));

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start_defaults(GTK_BOX(hbox),
	                        generic_button("Play", play_callback));
	gtk_box_pack_start_defaults(GTK_BOX(hbox),
	                        generic_button("Cache",
				               start_caching_callback));

	info->loop_toggle = (GtkToggleButton*)gtk_toggle_button_new_with_label(
						"Looping");
	gtk_widget_show(GTK_WIDGET(info->loop_toggle));
	gtk_box_pack_start_defaults(GTK_BOX(hbox),
	                            GTK_WIDGET(info->loop_toggle));
	gtk_widget_show(hbox);
	gtk_box_pack_start_defaults(GTK_BOX(vbox), GTK_WIDGET(hbox));

	info->button_box = gtk_vbox_new(TRUE, 0);
	gtk_widget_show(info->button_box);
	gtk_box_pack_start_defaults(GTK_BOX(vbox),
	                            GTK_WIDGET(info->button_box));
	gtk_widget_show(vbox);
	gtk_container_add(GTK_CONTAINER(info->top_level), GTK_WIDGET(vbox));
	gtk_widget_show(GTK_WIDGET(info->top_level));
}
