/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
/* DESCRIPTION
 * Sequencer test.
 *
 * Function: demo_sequencer
 */

/* TODO:
      dnd
      delete subsample
      delete buffer
      move subsample
      configure subsample
      gridding
      save/load
*/

#include <gtk/gtk.h>
#include <gdam/gdamsample.h>
#include <gdam/gdamchannel.h>
#include <stdlib.h>
#include <stdio.h>

typedef struct _DemoInfo DemoInfo;
typedef struct _BufferInfo BufferInfo;
typedef struct _SubsampleInfo SubsampleInfo;
typedef struct _ChooserInfo ChooserInfo;

#define	DEMO_BY_GAIN		300
#define	DEMO_BY_RESAMPLE	301
#define	DEMO_BY_ORDER		302

#define MAX_GAIN                (2*GdamFixed_ONE)
#define MAX_RESAMPLE            (2*GdamFixed_ONE)

struct _BufferInfo {
	int		buffer_id;
	GtkWidget*      toplevel;
	char*		rep;
	/* Offset of the center of the string in the current font
	 * (from the baseline, where it is drawn ... ) */
	int		rep_offset_x, rep_offset_y;
	DemoInfo*	demo_info;
	char*		filename;		/* or "description" */
};

struct _SubsampleInfo {
	BufferInfo*	buffer;
	int		start_time;
	GdamFixed	resample;
	GdamFixed	gain;

	/* Maintained by calls to `recompute_order'. */
	int		ordered_y;
};


struct _DemoInfo {
	GList*		buffers;
	GList*		subsamples;

	GdamChannel*	channel;
	guint		mixer_id;

	guint		sequence_id;

	/* Number of samples in this sequence. */
	int		sequence_length;

	/* The moving beat area. */
	GtkWidget*	sequence_widget;

	/* Holds all the subsample-types... */
	GtkBox*		samples_box;

	/* A widget to allow you to add samples to the sample box. */
	GtkWidget*	sample_adder;

	/* DEMO_BY_GAIN, DEMO_BY_RESAMPLE, DEMO_BY_ORDER */
	int		display_mode;

	/* The drawing area for the samples... */
	GtkWidget*	drawing_area;

	/* Font for the floating subsamples. */
	GdkFont*	font;

	/* Rendering information. */
	int		circle_radius;
	GdkGC*		text_gc;
	GdkGC*		circle_gc;
	GdkGC*		grid_gc;
	GdkGC*		background;

	/* Grid line locations. */
	int		num_time_grid;
	int*		time_grid;
	int		num_resample_grid;
	int*		resample_grid;
	int		num_gain_grid;
	int*		gain_grid;

	/* The new type of sample to add when clicked. */
	BufferInfo*	default_buffer;

	/* the gain/resample of new sources. */
	GdamFixed	default_resample, default_gain;

	/* doublebuffering */
	GdkPixmap*	backing_pixmap;

	/* If the user is dragging a subsample around,
	 * that is indicated here. */
	SubsampleInfo*	dragging;

	/* Offset from the center of the rendering where the user clicked. */
	int		drag_click_offset_x, drag_click_offset_y;

	/* Height and width of the drawing area. */
	int		draw_width, draw_height;


	/* Members used by the chooser widget. */
	ChooserInfo*	chooser_info;
	GtkLabel*	current_sample_label;
};
struct _ChooserInfo {
	GtkWidget* window;
	GtkWidget* file_button;
	GtkLabel* preview_label;
	GtkFileSelection* file_selection;
	gboolean is_visible;
	DemoInfo* demo_info;
};


typedef struct _RenderInfo {
	int x,y,w,h;
	DemoInfo* info;
} RenderInfo;

static int get_y_position(SubsampleInfo* sample, DemoInfo* info)
{
	switch (info->display_mode) {
		case DEMO_BY_GAIN:
			return info->draw_height
			     - sample->gain * info->draw_height / MAX_GAIN;
		case DEMO_BY_RESAMPLE:
			return info->draw_height
			     - sample->resample*info->draw_height/MAX_RESAMPLE;
		case DEMO_BY_ORDER:
			return sample->ordered_y;
		default: g_assert_not_reached();
	}
	return 0;
}

static void get_subsample_position(SubsampleInfo* sample, DemoInfo* info,
                                   int* out_x, int* out_y)
{
	*out_x = info->draw_width * sample->start_time / info->sequence_length;
	*out_y = get_y_position(sample, info);
}
static void recompute_order(DemoInfo* info)
{
	int count = g_list_length(info->subsamples);
	int index = 0;
	GList* at;
	for (at = info->subsamples; at; at = at->next) {
		SubsampleInfo* sub = (SubsampleInfo*)(at->data);
		sub->ordered_y = index * info->draw_height / count;
		index++;
	}
}
static void build_starttime_grid(DemoInfo* info, int num)
{
	int i;
	int* times = g_new(int, num);
	for (i = 0; i < num; i++)
		times[i] = info->sequence_length * i / num;
	info->time_grid = times;
	info->num_time_grid = num;
}

static void alloc_colors(DemoInfo* info)
{
	int state = GTK_WIDGET_STATE(info->drawing_area);
	/** HACK!!! */
	info->circle_gc = info->drawing_area->style->bg_gc[state];
	info->text_gc = info->drawing_area->style->bg_gc[state];
	info->background = info->drawing_area->style->fg_gc[state];
	info->font = info->drawing_area->style->font;

}
static void update_sequence_pattern(DemoInfo* info)
{
	int* elements;
	int n = g_list_length(info->subsamples), i = 0;
	GList* at = info->subsamples;
	elements = (int*)alloca(sizeof(int) * n * 6);
	recompute_order(info);
	while (at) {
		SubsampleInfo* subinfo = (SubsampleInfo*)(at->data);
		elements[i++] = subinfo->buffer->buffer_id;
		elements[i++] = subinfo->start_time;
		elements[i++] = subinfo->resample;	/* resample */
		elements[i++] = subinfo->gain;
		elements[i++] = 0;
		elements[i++] = -1;
		at = at->next;
	}
	{ int j; fprintf(stderr, "seqelements:"); for (j = 0; j < i; j++)
	fprintf(stderr, " %d", elements[j]); fprintf(stderr, ".\n"); }
	gdam_channel_source_set_arg(info->channel,
	                            info->sequence_id,
				    "elements",
				    GDAM_ARG_TYPE_INT_ARRAY,
				    i,
				    elements);
}

static void render_subsample(SubsampleInfo* subsample_info,
                             RenderInfo* render_info)
{
	GdkPixmap* pixmap = render_info->info->backing_pixmap;
	int x, y, radius;
	DemoInfo* demo_info = render_info->info;
	get_subsample_position(subsample_info, demo_info, &x, &y);
	radius = demo_info->circle_radius;

	/*
	 * Does the box around x/y touch the exposed area?
	 */
	if (x - radius > render_info->x + render_info->w
	 || x + radius < render_info->x
	 || y - radius > render_info->y + render_info->h
	 || y + radius < render_info->y)
	 	return;

	gdk_draw_arc(pixmap,
	             render_info->info->circle_gc,
		     FALSE,
		     x - radius + 1,
		     y - radius + 1,
		     radius * 2 - 1,
		     radius * 2 - 1,
		     0,
		     360*64);
	gdk_draw_string(pixmap,
	                demo_info->font,
		        demo_info->text_gc,
		        x - subsample_info->buffer->rep_offset_x,
		        y - subsample_info->buffer->rep_offset_y,
		        subsample_info->buffer->rep);
}

static void copy_backing_area(DemoInfo* info, int x, int y, int w, int h)
{
        gdk_draw_pixmap(info->drawing_area->window,
                        info->drawing_area->style
				->fg_gc[GTK_WIDGET_STATE(info->drawing_area)],
                        info->backing_pixmap,
                        x, y, x, y, w, h);
}
static void redraw_area(DemoInfo* info, int x, int y, int w, int h)
{
	RenderInfo render_info;
	render_info.x = x;
	render_info.y = y;
	render_info.w = w;
	render_info.h = h;
	render_info.info = info;
	if (!info->circle_gc)
		alloc_colors(info);
	gdk_draw_rectangle(info->backing_pixmap,
	                   info->background,
		           TRUE,
		           x,y,w,h);
	g_list_foreach(info->subsamples,
	               (GFunc)render_subsample,
		       &render_info);
	copy_backing_area(info, x, y, w, h);
}
static char* alloc_rep(DemoInfo* info)
{
	static int last_rep = -1;
	GList* list;
	for (list = info->buffers; list; list = list->next) {
		BufferInfo* bi = (BufferInfo*)(list->data);
		int i = atoi(bi->rep);
		if (i > last_rep)
			last_rep = i;
	}
	return g_strdup_printf("%d", last_rep+1);
}

#define DIST(a,b)  ( ((a) < (b)) ? ((b) - (a)) : ((a) - (b)) )
static void match_line(int* adjust, int num, const int* vals)
{
	int closest;
	int closest_dist;
	int saught = *adjust, i;
	if (num <= 0)
		return;
	g_return_if_fail(vals != NULL);

	closest = 0;
	closest_dist = DIST(saught, vals[0]);
	for (i = 1; i < num; i++) {
		int dist = DIST(saught, vals[i]);
		if (dist < closest_dist) {
			closest = i;
			closest_dist = dist;
		}
	}
	*adjust = vals[closest];
}

static void
snap_to_grid(DemoInfo* info, int x, int y, SubsampleInfo* sub_info)
{
	int start_time = x * info->sequence_length / info->draw_width;
	match_line(&start_time, info->num_time_grid, info->time_grid);
	sub_info->start_time = start_time;

	/* For the purposes of the following computations
	 * and anything involving resample/gain we invert
	 * y, so it corresponds to sheet music, math, etc,
	 * instead of computers. */
	y = info->draw_height - y;
	switch (info->display_mode) {
		case DEMO_BY_GAIN: {
			int gain;
			gain = y * MAX_GAIN / info->draw_height;
			match_line(&gain,
				   info->num_gain_grid,
			           info->gain_grid);
			sub_info->gain = gain;
			break;
		}
		case DEMO_BY_RESAMPLE: {
			int resample;
			resample = y * MAX_RESAMPLE / info->draw_height;
			match_line(&resample,
				   info->num_resample_grid,
			           info->resample_grid);
			sub_info->resample = resample;
			break;
		}
		case DEMO_BY_ORDER:
			/* never has grid lines */
			break;
		default: g_assert_not_reached();
	}
}

static gint
sequence_edit_expose(GtkDrawingArea* area,
                     GdkEventExpose* expose_event,
		     DemoInfo* info)
{
	redraw_area(info, 
		    expose_event->area.x, expose_event->area.y,
                    expose_event->area.width, expose_event->area.height);
	return 0;
}

GList* find_subcomponent_list_el(DemoInfo* info, int x, int y)
{
	GList* at;
	for (at = info->subsamples; at; at = at->next) {
		int pos_x, pos_y;
		SubsampleInfo* subsample = (SubsampleInfo*)at->data;
		get_subsample_position(subsample,
		                       info,
                                       &pos_x, &pos_y);
		if (pos_x - info->circle_radius <= x
		 && x <= pos_x + info->circle_radius
		 && pos_y - info->circle_radius <= y
		 && y <= pos_y + info->circle_radius) {
		 	return at;
		}
	}
	return NULL;
}
SubsampleInfo* find_subcomponent(DemoInfo* info, int x, int y)
{
	GList* out = find_subcomponent_list_el(info, x, y);
	return (SubsampleInfo*) (out ? out->data : NULL);
}

static void begin_dragging(DemoInfo* demo_info, SubsampleInfo* sample_info)
{
	demo_info->dragging = sample_info;
}

static void delete_subinfo(DemoInfo* info, SubsampleInfo* sample_info)
{
	int sample_x, sample_y, radius = info->circle_radius;
	info->subsamples = g_list_remove(info->subsamples, sample_info);
	if (info->dragging == sample_info)
		info->dragging = NULL;
	get_subsample_position(sample_info, info, &sample_x, &sample_y);
	redraw_area(info, sample_x - radius - 1, sample_y - radius - 1,
	                  2 * radius + 3, 2 * radius + 3);
	update_sequence_pattern(info);
	g_free(sample_info);
}

static gint
sequence_edit_button_press(GtkDrawingArea* area,
                           GdkEventButton* button_event,
		           DemoInfo* info)
{
	int press_x = button_event->x;
	int press_y = button_event->y;
	int width, height;
	SubsampleInfo* sample_info;
	g_return_val_if_fail(info->dragging == NULL, 0);
	if ((sample_info=find_subcomponent(info,
	                                   press_x, press_y)) != NULL) {
		int pos_x, pos_y;
		if (button_event->button == 3) {
			delete_subinfo(info, sample_info);
			return 0;
		}
		get_subsample_position(sample_info, info, &pos_x, &pos_y);
		info->drag_click_offset_x = press_x - pos_x;
		info->drag_click_offset_y = press_y - pos_y;
		begin_dragging(info, sample_info);
		return 0;
	}
	if (!info->default_buffer) {
		g_message("no sample has been chosen -- cannot add");
		return 0;
	}

	width = info->draw_width;
	height = info->draw_height;
	sample_info = g_new(SubsampleInfo, 1);
	sample_info->buffer = info->default_buffer;
	sample_info->start_time = press_x * info->sequence_length / width;
	sample_info->resample = info->default_resample;
	sample_info->gain = info->default_gain;
	sample_info->ordered_y = 0;
	info->subsamples = g_list_append(info->subsamples, sample_info);

	snap_to_grid(info, press_x, press_y, sample_info);
	get_subsample_position(sample_info, info, &press_x, &press_y);
	redraw_area(info,
	            press_x - info->circle_radius,
	            press_y - info->circle_radius,
	            info->circle_radius * 2 + 1,
	            info->circle_radius * 2 + 1);
	update_sequence_pattern(info);
	return 0;
}

static gint
sequence_edit_button_release(GtkDrawingArea* area,
                             GdkEventButton* button_event,
		             DemoInfo* info)
{
	int new_start_time;
	int release_x;
	int release_y;
	int old_x, old_y;
	int new_x, new_y;
	int radius = info->circle_radius;
	SubsampleInfo* dragging = info->dragging;
	if (!dragging)
		return 0;
	info->dragging = NULL;
	release_x = button_event->x - info->drag_click_offset_x;
	release_y = button_event->y - info->drag_click_offset_y;
	new_start_time = release_x * info->sequence_length / info->draw_width;

	get_subsample_position(dragging, info, &old_x, &old_y);
	snap_to_grid(info, release_x, release_y, dragging);
	get_subsample_position(dragging, info, &new_x, &new_y);

	redraw_area(info, 
	            old_x - radius - 1, old_y - radius - 1,
		    radius * 2 + 3, radius * 2 + 3);
	redraw_area(info, 
	            new_x - radius - 1, new_y - radius - 1,
		    radius * 2 + 3, radius * 2 + 3);

	update_sequence_pattern(info);
	return 0;
}

static gint
sequence_edit_configure(GtkWidget* widget,
                        GdkEventConfigure* configure_event,
		        DemoInfo* info)
{
	if (info->backing_pixmap)
		gdk_pixmap_unref(info->backing_pixmap);
	info->backing_pixmap = gdk_pixmap_new(widget->window,
	                              widget->allocation.width,
	                              widget->allocation.height,
				      -1);
	info->draw_width = widget->allocation.width;
	info->draw_height = widget->allocation.height;
	return TRUE;
}
static void sort_by_resample(DemoInfo* info)
{
	info->display_mode = DEMO_BY_RESAMPLE;
	redraw_area(info, 0, 0, info->draw_width, info->draw_height);
}
static void sort_by_gain(DemoInfo* info)
{
	info->display_mode = DEMO_BY_GAIN;
	redraw_area(info, 0, 0, info->draw_width, info->draw_height);
}
static void sort_by_order(DemoInfo* info)
{
	info->display_mode = DEMO_BY_ORDER;
	redraw_area(info, 0, 0, info->draw_width, info->draw_height);
}
static void 
get_string_offset(GdkFont* font, const char* string,
                  int* offset_x, int* offset_y)
{
	int lbearing, rbearing, width, ascent, descent;
	gdk_string_extents(font, string,
	                   &lbearing, &rbearing,
			   &width, &ascent, &descent);
	*offset_x = (lbearing+rbearing) / 2;
	*offset_y = 0;		/* at baseline?? should be centered?? */
}

GtkWidget*      make_sequence_edit_widget(DemoInfo*   info)
{
	GtkWidget* top_bar;
	GtkBox* vbox;
	GtkWidget* rv;
	GtkWidget* drawing_area;

	g_return_val_if_fail(info->sequence_widget == NULL, NULL);

	vbox = GTK_BOX(gtk_vbox_new(FALSE, 2));
	top_bar = gtk_hbox_new(FALSE, 2);

	info->current_sample_label = (GtkLabel*)gtk_label_new("");
	gtk_widget_show(GTK_WIDGET(info->current_sample_label));
	gtk_box_pack_start_defaults(GTK_BOX(top_bar),
	                            (GtkWidget*)info->current_sample_label);


	{ GtkBox* button_box = GTK_BOX(gtk_vbox_new(TRUE, 1));
	  GSList* list = NULL;
	  GtkWidget* radio;

	  radio = gtk_radio_button_new_with_label(list, "Resample");
	  gtk_widget_show(radio);
	  gtk_box_pack_start_defaults(button_box, radio);
	  gtk_signal_connect_object(GTK_OBJECT(radio),
	                            "clicked",
			            GTK_SIGNAL_FUNC(sort_by_resample),
				    /* May *not* use GTK_OBJECT()
				     * for the next cast.  See docs */
				    (GtkObject*)info);
	  list = gtk_radio_button_group(GTK_RADIO_BUTTON(radio));
	                                    
	  radio = gtk_radio_button_new_with_label(list, "Volume");
	  gtk_widget_show(radio);
	  gtk_box_pack_start_defaults(button_box, radio);
	  gtk_signal_connect_object(GTK_OBJECT(radio),
	                            "clicked",
			            GTK_SIGNAL_FUNC(sort_by_gain),
				    /* May *not* use GTK_OBJECT()
				     * for the next cast.  See docs */
				    (GtkObject*)info);
	  list = gtk_radio_button_group(GTK_RADIO_BUTTON(radio));

	  radio = gtk_radio_button_new_with_label(list, "Spaced");
	  gtk_widget_show(radio);
	  gtk_box_pack_start_defaults(button_box, radio);
	  gtk_signal_connect_object(GTK_OBJECT(radio),
	                            "clicked",
			            GTK_SIGNAL_FUNC(sort_by_order),
				    /* May *not* use GTK_OBJECT()
				     * for the next cast.  See docs */
				    (GtkObject*)info);
	  list = gtk_radio_button_group(GTK_RADIO_BUTTON(radio));

	  gtk_widget_show(GTK_WIDGET(button_box));
	  gtk_box_pack_start_defaults(GTK_BOX(top_bar),
	                                      GTK_WIDGET(button_box));
	}
	gtk_widget_show(GTK_WIDGET(top_bar));
	gtk_box_pack_start_defaults(vbox, GTK_WIDGET(top_bar));

	info->drawing_area = drawing_area = gtk_drawing_area_new();
	gtk_drawing_area_size(GTK_DRAWING_AREA(drawing_area), 800, 300);
	gtk_signal_connect(GTK_OBJECT(drawing_area),
	                   "button_press_event",
			   GTK_SIGNAL_FUNC(sequence_edit_button_press),
			   info);
	gtk_signal_connect(GTK_OBJECT(drawing_area),
	                   "button_release_event",
			   GTK_SIGNAL_FUNC(sequence_edit_button_release),
			   info);
	gtk_signal_connect(GTK_OBJECT(drawing_area),
	                   "expose_event",
			   GTK_SIGNAL_FUNC(sequence_edit_expose),
			   info);
	gtk_signal_connect(GTK_OBJECT(drawing_area),
	                   "configure_event",
			   GTK_SIGNAL_FUNC(sequence_edit_configure),
			   info);
	gtk_widget_show(drawing_area);
	gtk_widget_add_events(drawing_area, GDK_EXPOSURE_MASK
                                          | GDK_LEAVE_NOTIFY_MASK
                                          | GDK_BUTTON_PRESS_MASK
                                          | GDK_BUTTON_RELEASE_MASK
                                          | GDK_POINTER_MOTION_MASK
                                          | GDK_POINTER_MOTION_HINT_MASK);
	gtk_box_pack_start_defaults(vbox, drawing_area);

	/* a vbox of samples available */
	info->samples_box = GTK_BOX(gtk_vbox_new(FALSE, 2));
	gtk_widget_show(GTK_WIDGET(info->samples_box));
	gtk_box_pack_start_defaults(vbox, GTK_WIDGET(info->samples_box));

	rv = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_widget_show(GTK_WIDGET(vbox));
	gtk_container_add(GTK_CONTAINER(rv), GTK_WIDGET(vbox));
	gtk_widget_show(rv);
	return rv;
}

static void default_selected(GtkButton* button, BufferInfo* buffer_info)
{
	buffer_info->demo_info->default_buffer = buffer_info;
}

static void delete_buffer(GtkButton* button, BufferInfo* buffer_info)
{
}

static void done_caching(gpointer _buffer_info)
{
	BufferInfo* buffer_info = (BufferInfo*)_buffer_info;
	DemoInfo* demo_info = buffer_info->demo_info;
	GtkWidget* button;
	GtkWidget* destroy_button;
	char* label_text;

	g_return_if_fail(buffer_info != NULL && demo_info != NULL);
	buffer_info->rep = alloc_rep(demo_info);
	demo_info->buffers = g_list_append(demo_info->buffers, buffer_info);
	get_string_offset(demo_info->drawing_area->style->font,
	                  buffer_info->rep,
			  &buffer_info->rep_offset_x,
			  &buffer_info->rep_offset_y);
	label_text = g_strdup_printf("%s -- %s",
	                             buffer_info->rep,
				     buffer_info->filename);
	button = gtk_button_new_with_label(label_text);
	g_free(label_text);
	destroy_button = gtk_button_new_with_label("X");
	buffer_info->toplevel = gtk_hbox_new(FALSE, 2);
	gtk_signal_connect(GTK_OBJECT(button),
	                   "clicked",
			   GTK_SIGNAL_FUNC(default_selected),
			   buffer_info);
	gtk_signal_connect(GTK_OBJECT(buffer_info->toplevel),
	                   "destroy",
			   GTK_SIGNAL_FUNC(delete_buffer),
			   buffer_info);
	gtk_signal_connect_object(GTK_OBJECT(destroy_button),
	                          "clicked",
				  GTK_SIGNAL_FUNC(gtk_widget_destroy),
				  GTK_OBJECT(buffer_info->toplevel));
	gtk_box_pack_start_defaults(GTK_BOX(buffer_info->toplevel),
	                            GTK_WIDGET(button));
	gtk_box_pack_start_defaults(GTK_BOX(buffer_info->toplevel),
	                            GTK_WIDGET(destroy_button));
	gtk_widget_show(destroy_button);
	gtk_widget_show(button);
	gtk_widget_show(buffer_info->toplevel);
	gtk_box_pack_start_defaults(demo_info->samples_box,
	                            GTK_WIDGET(buffer_info->toplevel));
}
	
static void start_caching(DemoInfo* demo, int source_id,
                          const char* description)
{
	BufferInfo* buffer_info;
	buffer_info = g_new(BufferInfo, 1);
	buffer_info->buffer_id = gdam_channel_start_caching(demo->channel,
	                                                    source_id,
					                    done_caching,
					                    buffer_info);
	g_assert(buffer_info->buffer_id != 0);
	buffer_info->toplevel = NULL;
	buffer_info->rep = NULL;
	buffer_info->filename = g_strdup(description);
	buffer_info->rep_offset_x = buffer_info->rep_offset_y = 0;
	buffer_info->demo_info = demo;
}

static void file_chosen(ChooserInfo* chooser)
{
	char* filename;
	int id;
	filename = gtk_file_selection_get_filename(chooser->file_selection);

	g_return_if_fail(filename != NULL);
	id = gdam_channel_create_source_generic(chooser->demo_info->channel,
	                                 "GdamWav",
					 2,
					 "filename",
					 GDAM_ARG_TYPE_STRING,
					 filename,
					 "sequence_length",
					 GDAM_ARG_TYPE_INT,
					 44100);
	start_caching(chooser->demo_info, id, filename);
}

static void toggle_file_selection(GtkWidget* button, ChooserInfo* chooser)
{
	if (chooser->file_selection == NULL) {
		chooser->file_selection = GTK_FILE_SELECTION(
				gtk_file_selection_new("Pick wav file"));
		gtk_signal_connect_object(
				GTK_OBJECT(chooser->file_selection->ok_button),
		                   "clicked",
				   GTK_SIGNAL_FUNC(file_chosen),
				   (GtkObject*)chooser);
		chooser->is_visible = FALSE;
	}
	if (chooser->is_visible)
		gtk_widget_hide(GTK_WIDGET(chooser->file_selection));
	else
		gtk_widget_show(GTK_WIDGET(chooser->file_selection));

	chooser->is_visible = ! chooser->is_visible;
}


GtkWidget* make_sequence_chooser(DemoInfo* info)
{
	ChooserInfo* chooser_info;
	//char* fname = g_getenv("DEMOFILE");
	GtkBox* hbox;
	chooser_info = g_new(ChooserInfo, 1);
	chooser_info->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	chooser_info->file_button = gtk_button_new_with_label("File...");
	/*chooser_info->preview_label = (GtkLabel*)
					gtk_label_new(fname?fname:"...");*/
	chooser_info->file_selection = NULL;
	chooser_info->is_visible = FALSE;
	chooser_info->demo_info = info;
	gtk_signal_connect(GTK_OBJECT(chooser_info->file_button),
	                   "clicked",
			   GTK_SIGNAL_FUNC(toggle_file_selection),
			   chooser_info);

	hbox = GTK_BOX(gtk_hbox_new(FALSE, 2));
	gtk_widget_show(GTK_WIDGET(chooser_info->file_button));
	gtk_box_pack_start_defaults(hbox,
	                            GTK_WIDGET(chooser_info->file_button));
	/*gtk_widget_show(GTK_WIDGET(chooser_info->preview_label));
	gtk_box_pack_start_defaults(hbox,
	                            GTK_WIDGET(chooser_info->preview_label));*/
	gtk_widget_show(GTK_WIDGET(hbox));
	gtk_container_add(GTK_CONTAINER(chooser_info->window),
	                  GTK_WIDGET(hbox));
	gtk_widget_show(GTK_WIDGET(chooser_info->window));
	return chooser_info->window;
}

DemoInfo*  make_complete_demo_info(GdamChannel* channel, guint mixer_id)
{
	DemoInfo* info;
	info = g_new0(DemoInfo, 1);
	info->channel = channel;
	info->mixer_id = mixer_id;
	info->buffers = NULL;
	info->default_resample = info->default_gain = GdamFixed_ONE;
	info->subsamples = NULL;
	info->samples_box = NULL;
	info->sequence_widget = NULL;
	info->sample_adder = NULL;
	info->sequence_length = 44100;
	info->display_mode = DEMO_BY_GAIN;
	info->circle_radius = 24;
	info->text_gc = info->circle_gc = NULL;
	info->dragging = NULL;
	info->draw_width = 600;
	info->draw_height = 300;
	info->backing_pixmap = NULL;

	build_starttime_grid(info, 16);

	/* I basically assume (in the form of g_return_if_fails)
	 * that chooser's callbacks will have `sequence_widget',
	 * so the order of construction *is* important. */
	info->sequence_widget = make_sequence_edit_widget(info);
	info->sample_adder = make_sequence_chooser(info);

	return info;
}


void demo_sequencer(GdamChannel*           channel,
                    guint                  mixer_id)
{
	DemoInfo* info;
	info = make_complete_demo_info(channel, mixer_id);
	info->sequence_id = gdam_channel_create_source_generic(
				channel,
				"GdamSequencer",
				1,
				"sequence_length",
				GDAM_ARG_TYPE_INT,
				44100);
	g_assert(info->sequence_id != 0);
	gdam_channel_mix_add(channel, mixer_id, info->sequence_id, -1);
}
