/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamsource.h"
#include "gdammix.h"
#include "gdamdispatch.h"
#include "gdamnamespace.h"
#include <gdam/gdam-protocol.h>
#include <assert.h>
#include <stdio.h>


#define DEFAULT_MIX_BUFFER	2048

/* Please look at the casts to this function below before
 * changing its prototype. */
typedef void (*SampleTransformFunc)(GdamSample*       dest,
                                    const GdamSample* src,
				    int               num,
				    gpointer          data);
	
static GdamMixElement* gdam_mix_connect_internal(GdamMix*    mix,
                                                 GdamSource* add,
						 gint64      when);

static void        gdam_mix_class_init       (GdamSourceClass* klass);
static void        gdam_mix_init             (GdamMix*         mix);
static void        gdam_mix_destroy          (GdamSource*      source);
static GdamBuffer* gdam_mix_interpret_command(GdamSource*      source,
					      GdamBuffer*      command,
					      gboolean*        was_ok);
static gboolean    gdam_mix_is_ready_to_read (GdamSource*      source);
static int	   gdam_mix_read             (GdamSource*      input,
					      GdamSample*      samples,
					      int              max_read);
static GdamTime    gdam_mix_get_position     (GdamSource*      source);
static void        gdam_mix_get_latency_info (GdamSource*      source_mix,
                                              GdamSource*      sub,
				              GdamLatencyInfo* info);
static void        gdam_mix_miss_data        (GdamSource*      input,
					      int              lost);
static gboolean    gdam_mix_detach_child     (GdamSource*      mix,
                                              GdamSource*      sub_source);
static void        gdam_mix_get_subsources   (GdamSource*      source,
                                              int*             num_subs,
					      GdamSource***    subs);

static void        delete_element            (GdamMix* mix,
                                              GdamMixElement* element);

/* TODO:  on each read, sort the elements by the amount of
 *        blank space they have at the beginning... we should
 *        gdam_sample_zero only if there are no sources ready to read... */

static GtkObjectClass* parent_class = NULL;


static void gdam_mix_init(GdamMix* mix)
{
	mix->first_source      = 0;
	mix->last_source       = 0;
	mix->buffer_size       = DEFAULT_MIX_BUFFER;
	mix->buffer            = g_new(GdamSample, mix->buffer_size);
	mix->buffer_data_start = 0;
	mix->tmp_buf           = g_new(GdamSample, mix->buffer_size);
	mix->samples_output    = 0;
	mix->samples_remaining = -1;
	mix->mode              = GDAM_MIX_ETERNAL;
}

#ifdef gdam_sample_copy
#undef gdam_sample_copy
#endif

void gdam_mix_arg_set_mixed(GtkObject* object, GdamArg* arg)
{
	GdamMix* mix = (GdamMix*)object;
	gint64* array = GDAM_ARG_INT64_ARRAY(arg).array;
	int num_els = GDAM_ARG_INT64_ARRAY(arg).length, i;
	g_return_if_fail(num_els % 2 == 0);
	num_els /= 2;

	while (mix->first_source)
		delete_element(mix, mix->first_source);
	
	for (i = 0; i < num_els; i++) {
		gint64 position = (int)array[2*i+0];
		int source_id = (int)array[2*i+1];
		GdamSource* subsource = gdam_namespace_local_source(source_id);
		if (subsource == NULL) {
			gdam_debug("no source of id %d "
					"(needed in setting mixed)");
			continue;
		}
		gdam_mix_connect(mix, subsource, position);
	}
}

void gdam_mix_arg_get_mixed(GtkObject* object, GdamArg* arg)
{
	GdamMix* mix = GDAM_MIX(object);
	GdamMixElement* mix_element = mix->first_source;
	int num_subs = gdam_mix_num_sources(mix);
	gint64* rv = g_new(gint64, num_subs * 2);
	int i;
	for (i = 0; i < num_subs; i++) {
		rv[i*2 + 0] = mix_element->read_to;
		rv[i*2 + 1] = mix_element->sub_source->source_id;

		mix_element = mix_element->next;
	}
	GDAM_ARG_INT64_ARRAY(arg).length = num_subs * 2;
	GDAM_ARG_INT64_ARRAY(arg).array = rv;
}
static void gdam_mix_set_terminates(GtkObject* object, GdamArg* arg)
{
	gdam_debug("use of terminates is deprecated: use `mode=%s' instead",
		GDAM_ARG_INT(arg) ? "all_ended" : "eternal");
	GDAM_MIX(object)->mode = GDAM_ARG_INT(arg) ? GDAM_MIX_ALL_ENDED
	                                           : GDAM_MIX_ETERNAL;
}
static void gdam_mix_get_terminates(GtkObject* object, GdamArg* arg)
{
	GDAM_ARG_INT(arg) = ! (GDAM_MIX(object)->mode == GDAM_MIX_ETERNAL);
}

static void gdam_mix_set_mode(GtkObject* object, GdamArg* arg)
{
	const char* mode = GDAM_ARG_STRING(arg);
	GdamMix* mix = GDAM_MIX(object);
	if (strcmp(mode, "all_ended") == 0) {
		mix->mode = GDAM_MIX_ALL_ENDED;
	} else if (strcmp(mode, "eternal") == 0) {
		mix->mode = GDAM_MIX_ETERNAL;
	} else if (strcmp(mode, "first_ended") == 0) {
		mix->mode = GDAM_MIX_FIRST_ENDED;
	} else {
		gdam_debug("gdam_mix_set_mode: unknown mode %s", mode);
	}
}
static void gdam_mix_get_mode(GtkObject* object, GdamArg* arg)
{
	switch (GDAM_MIX(object)->mode) {
		case GDAM_MIX_ETERNAL:
			GDAM_ARG_STRING(arg) = g_strdup("eternal");
			break;
		case GDAM_MIX_ALL_ENDED:
			GDAM_ARG_STRING(arg) = g_strdup("all_ended");
			break;
		case GDAM_MIX_FIRST_ENDED:
			GDAM_ARG_STRING(arg) = g_strdup("first_ended");
			break;
		default:
			g_assert_not_reached();
	}
}

static void gdam_mix_arg_add_with_latency(GtkObject* object, GdamArg* arg)
{
	GdamMix* mix = GDAM_MIX(object);
	GdamTime when = mix->samples_output;
	GdamLatencyInfo latency;
	GdamSource* sub = gdam_namespace_local_source(GDAM_ARG_INT(arg));
	if (sub == NULL) {
		gdam_debug("add_with_latency(%d): source didn't exist",
			GDAM_ARG_INT(arg));
		return;
	}
	gdam_source_find_total_latency(sub, &latency);
	when += latency.latency;
	gdam_mix_connect_internal(mix, sub, when);
}

static void gdam_mix_arg_set_source_doesnt_miss(GtkObject* object, GdamArg* arg)
{
	GdamMixElement* element;
	GdamMix* mix = GDAM_MIX(object);
	GdamSource* sub = gdam_namespace_local_source(GDAM_ARG_INT(arg));
	for (element = mix->first_source;
	     element != NULL;
	     element = element->next) {
	     	if (element->sub_source == sub)
			break;
	}
	if (element)
		element->misses_data = FALSE;
}
static void gdam_mix_arg_set_source_does_miss(GtkObject* object, GdamArg* arg)
{
	GdamMixElement* element;
	GdamMix* mix = GDAM_MIX(object);
	GdamSource* sub = gdam_namespace_local_source(GDAM_ARG_INT(arg));
	for (element = mix->first_source;
	     element != NULL;
	     element = element->next) {
	     	if (element->sub_source == sub)
			break;
	}
	if (element)
		element->misses_data = TRUE;
}


static void gdam_mix_class_init(GdamSourceClass* klass)
{
	GtkObjectClass* object_class = (GtkObjectClass*)klass;
	GdamMixClass* mix_class = (GdamMixClass*)klass;
	if (!parent_class)
		parent_class = gtk_type_class(GDAM_TYPE_SOURCE);
	klass->read = gdam_mix_read;
	klass->miss_data = gdam_mix_miss_data;
	klass->interpret_command = gdam_mix_interpret_command;
	klass->is_ready_to_read = gdam_mix_is_ready_to_read;
	klass->detach_child = gdam_mix_detach_child;
	klass->get_latency_info = gdam_mix_get_latency_info;
	klass->get_position = gdam_mix_get_position;
	klass->get_subsources = gdam_mix_get_subsources;
	mix_class->initial_copy = (SampleTransformFunc)gdam_sample_copy;
	mix_class->augment = (SampleTransformFunc)gdam_sample_addto;
	object_class->destroy = (void(*)(GtkObject*)) gdam_mix_destroy;

	gdam_arg_register(object_class,
	                  "mixed",
			  GDAM_ARG_TYPE_INT64_ARRAY,
			  gdam_mix_arg_set_mixed,
			  gdam_mix_arg_get_mixed,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "terminates",
			  GDAM_ARG_TYPE_INT,
			  gdam_mix_set_terminates,
			  gdam_mix_get_terminates,
			  0);
	gdam_arg_register(object_class,
	                  "source_does_miss",
			  GDAM_ARG_TYPE_INT,
			  gdam_mix_arg_set_source_does_miss,
			  NULL,
			  0);
	gdam_arg_register(object_class,
	                  "source_doesnt_miss",
			  GDAM_ARG_TYPE_INT,
			  gdam_mix_arg_set_source_doesnt_miss,
			  NULL,
			  0);
	gdam_arg_register(object_class,
	                  "add_with_latency",
			  GDAM_ARG_TYPE_INT,
			  gdam_mix_arg_add_with_latency,
			  NULL,
			  0);
	gdam_arg_register(object_class,
	                  "mode",
			  GDAM_ARG_TYPE_STRING,
			  gdam_mix_set_mode,
			  gdam_mix_get_mode,
			  GDAM_ARG_VITAL);
}

static void delete_element(GdamMix* mix, GdamMixElement* element)
{
	gdam_debug("mix_delete_element: element=%p; source=%p; source_id=%d",
		element, element->sub_source, element->sub_source->source_id);
	element->sub_source->parent = NULL;
	if (element->name)
		g_hash_table_remove(mix->names, element->name);
	if (element->next)
		element->next->prev = element->prev;
	else
		mix->last_source = element->prev;
	if (element->prev)
		element->prev->next = element->next;
	else
		mix->first_source = element->next;
	gtk_object_unref((GtkObject*)(element->sub_source));
	g_free(element);
}

/* gdam_mix_subsource read calls ready_to_read on subsources,
   so it can deal with any situation... in fact, that unifies the places
   we have to deal with it...

   this might not be great if algorithms ever called gdam_mix_is_ready_to_read
   explorationally, but it suffices for now. */
static gboolean gdam_mix_is_ready_to_read(GdamSource* source)
{
	return TRUE;

	/*GdamMixElement* element;

	for (element = mix->first_source;
	     element != NULL;
	     element = element->next) {
	     	if (element->started <= mix->samples_output
		 && element->read_to > mix->samples_output)
		 	return TRUE;
		if (element->started > mix->samples_output)
			continue;
		
	     ...

	}*/
}
	
static void mix_zero_to(GdamMix* mix, GdamTime zero_to)
{
	int num_zero = (int)(zero_to - mix->cleared_to);
	int zero_index = (int)(mix->cleared_to % mix->buffer_size);
	if (num_zero > mix->buffer_size) {
		gdam_sample_zero(mix->buffer, mix->buffer_size);
	} else {
		int p1 = mix->buffer_size - zero_index;
		if (p1 > num_zero) {
			gdam_sample_zero(mix->buffer + zero_index, num_zero);
		} else {
			gdam_sample_zero(mix->buffer + zero_index, p1);
			gdam_sample_zero(mix->buffer, num_zero - p1);
		}
	}
	mix->cleared_to = zero_to;
}

/* Read from an element, returning whether or not to keep the source
 * around. */
static gboolean gdam_mix_subsource_read(GdamMix* mix, GdamMixElement* element)
{
	int max_read, r;
	int init_buffer_index;
	int num_clear;
	gint64 max_end_buf = mix->samples_output + mix->buffer_size;
	SampleTransformFunc add_func = NULL;
	SampleTransformFunc init_func = NULL;
	GdamMixClass* mix_class;
	gpointer element_data;

	if (max_end_buf <= element->read_to)
		return TRUE;
	max_read = (max_end_buf - element->read_to);
	g_assert(max_read <= mix->buffer_size);
	r = gdam_source_read(element->sub_source, mix->tmp_buf, max_read);
	if (r <= 0) {
		if (mix->mode == GDAM_MIX_FIRST_ENDED) {
			if (mix->samples_output <= element->read_to) {
				gint64 new_remaining;
				gint64 old_remaining = mix->samples_remaining;
				new_remaining = element->read_to
				              - mix->samples_output;
				if (old_remaining == -1
				 || new_remaining < old_remaining)
				 	mix->samples_remaining = new_remaining;
			} else
				mix->samples_remaining = 0;
		}
		return FALSE;
	}
	if (element->read_to > mix->cleared_to) {
		mix_zero_to(mix, element->read_to);
	}

	num_clear = (int)(element->read_to + r - mix->cleared_to);
	g_assert(num_clear <= r);
	if (num_clear < 0)
		num_clear = 0;
	else
		r -= num_clear;

	init_buffer_index = (int)(element->read_to % mix->buffer_size);

	element_data = element->source_info;

	mix_class = (GdamMixClass*) ((GtkObject*)mix)->klass;
	if (r > 0) {
		add_func = mix_class->augment;
		if (init_buffer_index + r > mix->buffer_size) {
			int second_part = mix->buffer_size - init_buffer_index;
			int first_part = r - second_part;
			assert(second_part > 0 && first_part > 0);
			add_func(mix->buffer + init_buffer_index,
					  mix->tmp_buf,
					  second_part,
					  element_data);
			add_func(mix->buffer,
					  mix->tmp_buf + second_part,
					  first_part,
					  element_data);
		} else {
			add_func(mix->buffer + init_buffer_index,
				 mix->tmp_buf,
				 r,
				 element_data);
		}
	}
	element->read_to += r;
	if (num_clear > 0) {
		init_buffer_index += r;
		init_buffer_index %= mix->buffer_size;
		init_func = mix_class->initial_copy;

		if (init_buffer_index + num_clear > mix->buffer_size) {
			int second_part = mix->buffer_size - init_buffer_index;
			int first_part = num_clear - second_part;
			assert(second_part > 0 && first_part > 0);
			init_func(mix->buffer + init_buffer_index,
					  mix->tmp_buf + r,
					  second_part,
					  element_data);
			init_func(mix->buffer,
					  mix->tmp_buf + second_part + r,
					  first_part,
					  element_data);
		} else {
			init_func(mix->buffer + init_buffer_index,
				 mix->tmp_buf + r,
				 num_clear,
				 element_data);
		}
		mix->cleared_to += num_clear;
		element->read_to += num_clear;
	}
	if (element->read_to > mix->last_read_to)
		mix->last_read_to = element->read_to;
	return TRUE;
}
/*
static inline void
gdam_sample_copy_and_zero(GdamSample* dst, GdamSample* src, int n)
{
	gdam_sample_copy(dst, src, n);
	gdam_sample_zero(src, n);
}
*/

/* Fill the mix buffer from each subsource.
 */
static int gdam_mix_read_one(GdamMix* mix, GdamSample* buf, int n)
{
	//GdamSource* source = (GdamSource*)mix;
	GdamMixElement* element, *next_element = 0;
	int to_read = n;
	GdamTime start_buffer_time = mix->samples_output;
	GdamTime end_buffer_time = start_buffer_time + mix->buffer_size;
	GdamTime end_read_time;

	/* Search through the sources and see if we can
	 * read on sources we want to read.  */
	for (element = mix->first_source;
	     element != NULL;
	     element = next_element) {
		element->flags = 0;
		next_element = element->next;
		/* if the element is too far in the future to read... */
		if (element->read_to >= end_buffer_time) {
			element->flags |= MIX_ELEMENT_IN_FUTURE;
			continue;
		}

	     	/* if we could read more, check if we want to read more. */
		if (gdam_source_is_ready_to_read(element->sub_source))
			if (!gdam_mix_subsource_read(mix, element)) {
				gdam_debug("gdam_mix_read: "
					"calling delete_element");
				delete_element(mix, element);
			}
	}

	/* Compute the amount to return.  (If sources were unable
	 * to read, and have no data buffered they will have the
	 * `miss_data' callback called.) */
	for (element = mix->first_source;
	     element != NULL;
	     element = element->next) {
		int data_available;
	     	if (element->flags & MIX_ELEMENT_IN_FUTURE)
			continue;
		data_available = (int)(element->read_to - start_buffer_time);
		if (data_available > 0
		 && data_available < to_read)
			to_read = data_available;
	}

	/* Call the miss_data virtuals, as needed. */
	end_read_time = start_buffer_time + to_read;
	for (element = mix->first_source;
	     element != NULL;
	     element = element->next) {
		int missed = (int)(end_read_time - element->read_to);
	     	if (missed > 0) {
			if (element->misses_data)
				gdam_source_miss_data(element->sub_source,
				                      missed);
			element->read_to = end_read_time;
		}
	}

	if (mix->cleared_to < end_read_time) {
		int num_to_clear = (int)(end_read_time - mix->cleared_to);
		if (num_to_clear > mix->buffer_size) {
			gdam_sample_zero(mix->buffer, mix->buffer_size);
		} else if (num_to_clear > 0) {
			int start_clear_index = (int)(mix->cleared_to
			                            % mix->buffer_size);
			int p1 = mix->buffer_size - start_clear_index;
			GdamSample* buf_start = mix->buffer+start_clear_index;
			if (p1 > num_to_clear) {
				gdam_sample_zero(buf_start, num_to_clear);
				mix->cleared_to += num_to_clear;
				//mix->buffer_data_start = 0;
			} else {
				int p2 = num_to_clear - p1;
				gdam_sample_zero(buf_start, p1);
				gdam_sample_zero(mix->buffer, p2);
				mix->cleared_to += num_to_clear;
				//mix->buffer_data_start = p2;
			}
		}
		mix->cleared_to = end_read_time;
	}
	/* Copy the data out of the buffer. */
	{
		int p1 = mix->buffer_size - mix->buffer_data_start;
		GdamSample* buf_start = mix->buffer+mix->buffer_data_start;
		g_assert(to_read > 0);
		if (p1 > to_read) {
			gdam_sample_copy(buf, buf_start, to_read);
			mix->buffer_data_start += to_read;
		} else {
			int p2 = to_read - p1;
			gdam_sample_copy(buf, buf_start, p1);
			gdam_sample_copy(buf + p1, mix->buffer, p2);
			mix->buffer_data_start = p2;
		}
	}
	if (mix->mode == GDAM_MIX_ALL_ENDED
	 && mix->first_source == NULL) {
		if (mix->last_read_to > mix->samples_output) {
			gint64 max_to_read;
			max_to_read = mix->last_read_to - mix->samples_output;
			if (max_to_read < to_read)
				to_read = (int)max_to_read;
		} else {
			to_read = 0;
		}
	} else if (mix->mode == GDAM_MIX_FIRST_ENDED) {
		if (mix->samples_remaining != -1
		 && mix->samples_remaining < to_read) {
			to_read = (int)mix->samples_remaining;
		}
	}
	g_assert(to_read >= 0);
	mix->samples_output += to_read;
	if (mix->samples_remaining >= 0)
		mix->samples_remaining -= to_read;
	return to_read;
}

int gdam_mix_read(GdamSource* source, GdamSample* buf, int n)
{
	int amount_read = 0;
	GdamMix* mix = GDAM_MIX(source);
	while (amount_read < n) {
		int i;
		i = gdam_mix_read_one(mix,
		                      buf + amount_read,
		                      n - amount_read);
		if (i < 0)
			return i;
		if (i == 0)
			return amount_read;
		g_assert(i <= n - amount_read);
		amount_read += i;
		if (i == mix->buffer_size) {
			gboolean had_one = FALSE, had_none = TRUE;
			GdamMixElement* at;
			for (at = mix->first_source; at; at = at->next) {
				GdamSource* sub = at->sub_source;
				had_none = FALSE;
				if (gdam_source_is_ready_to_read(sub)) {
					had_one = TRUE;
					break;
				}
			}
			if (!had_one && !had_none)
				return amount_read;
		} else {
			return amount_read;
		}
	}
	return amount_read;
}
		


static void gdam_mix_get_latency_info(GdamSource* source_mix,
                                      GdamSource* sub,
				      GdamLatencyInfo* info)
{
	GdamMixElement* element;
	GdamMix* mix = GDAM_MIX(source_mix);
	info->resample = GdamFixed_ONE;
	g_return_if_fail(mix != NULL);
	for (element = mix->first_source;
	     element != NULL;
	     element = element->next) {
	     	if (element->sub_source == sub) {
			int delta;
			delta = (int)(element->read_to - mix->samples_output);
			info->latency = delta;
			return;
		}
	}
	g_warning("gdam_mix_get_latency_info: unable to find subsource %p",sub);
	info->latency = 0;
	return;
}

/* TODO: fix these two functions (eliminate destruct) */
void mix_destruct(GdamSource* source)
{
	GdamMix* mix = (GdamMix*)source;
	while (mix->first_source)
		delete_element(mix, mix->first_source);
	g_free(mix->buffer);
	mix->buffer = 0;
	g_free(mix->tmp_buf);
	mix->tmp_buf = 0;
}

static void gdam_mix_destroy(GdamSource* source)
{
	GdamMix* mix = GDAM_MIX(source);
	while (mix->first_source)
		gdam_source_detach_child(source, mix->first_source->sub_source);
	g_free(mix->buffer);
	mix->buffer = NULL;
	if (mix->names) {
		g_hash_table_destroy(mix->names);
		mix->names = NULL;
	}
	(*parent_class->destroy)(GTK_OBJECT(source));
}

static GdamMixElement*
gdam_mix_connect_internal(GdamMix* mix, GdamSource* add, gint64 when)
{
	GdamMixElement* element;
	GdamMixElement* new_element;
	GtkObject* as_object = GTK_OBJECT(add);
	gint64 position;
	if (when == -1)
		when = mix->samples_output;

	g_return_val_if_fail(!gdam_source_is_ancestor(GDAM_SOURCE(mix), add),
	                     NULL);

	/* This increases the ref count of add. */
	gdam_source_orphan(add);

	for (element = mix->first_source;
	     element && element->read_to < when;
	     element = element->next)
	     	;

	new_element = g_new(GdamMixElement, 1);
	new_element->next = element;
	new_element->prev = element ? element->prev : mix->last_source;
	new_element->name = NULL;
	new_element->misses_data = TRUE;

	if (new_element->prev)
		new_element->prev->next = new_element;
	else
		mix->first_source = new_element;
	if (new_element->next)
		new_element->next->prev = new_element;
	else
		mix->last_source = new_element;
	position = gdam_source_get_position(add);
	gdam_debug("gdam_mix_connect: source %p (element %p) position %lld, "
	           "offset %lld", add, new_element, position, when - position);
	new_element->sub_source = add;
	gtk_object_sink(as_object);
	new_element->read_to = when;
	add->parent = (GdamSource*)mix;
	return new_element;
}
void gdam_mix_connect(GdamMix* mix, GdamSource* add, gint64 when)
{
	gdam_mix_connect_internal(mix, add, when);
}

void gdam_mix_connect_named(GdamMix* mix,
                            GdamSource* add,
                            GdamTime when,
			    const char* name)
{
	GdamMixElement* new_element;
	if (mix->names == NULL)
		mix->names = g_hash_table_new(g_str_hash, g_str_equal);
	else {
		GdamMixElement* old_element;
		old_element = (GdamMixElement*) g_hash_table_lookup(
							mix->names,
							name);
		g_hash_table_remove(mix->names, name);
		if (old_element) {
			when = old_element->read_to;
			delete_element(mix, old_element);
		}
	}
	if (add == NULL)
		return;
	new_element = gdam_mix_connect_internal(mix, add, when);
	if (new_element == NULL)
		return;
	new_element->name = g_strdup(name);
	g_hash_table_insert(mix->names, new_element->name, new_element);
}

GdamSource*    gdam_mix_get_named       (GdamMix*    mix,
                                         const char* name)
{
	GdamMixElement* element;
	if (mix->names == NULL)
		return NULL;
	element = (GdamMixElement*) g_hash_table_lookup(mix->names, name);
	return element ? element->sub_source : NULL;
}

static GdamTime    gdam_mix_get_position     (GdamSource*      source)
{
	return GDAM_MIX(source)->samples_output;
}
static void gdam_mix_miss_data(GdamSource* source, int lost)
{
	GdamMix* mix = GDAM_MIX(source);
	GdamTime new_read_to = mix->samples_output + lost;
	GdamMixElement* element;
	for (element = mix->first_source;
	     element;
	     element = element->next) {
	     	if (element->read_to < new_read_to) {
			int delta = (int)(new_read_to - element->read_to);
			gdam_source_miss_data(element->sub_source, delta);
			element->read_to = new_read_to;
	     	}
	}
	mix->samples_output = new_read_to;
	mix->buffer_data_start += lost;
	mix->buffer_data_start %= mix->buffer_size;
}

static gboolean gdam_mix_detach_child(GdamSource* super_source,
                                      GdamSource* sub_source)
{
	GdamMix* mix = GDAM_MIX(super_source);
	GdamMixElement* element;
	g_return_val_if_fail(mix != NULL, FALSE);
	for (element = mix->first_source;
	     element && element->sub_source != sub_source;
	     element = element->next)
	     	;
	g_return_val_if_fail(element != NULL, FALSE);
	if (!element)
		return FALSE;
	gdam_debug("gdam_mix_detach_child: calling delete_element");
	delete_element(mix, element);
	return TRUE;
}

static GdamBuffer* gdam_mix_interpret_command(GdamSource*      source,
					      GdamBuffer*      command,
					      gboolean*        was_ok)
{
	GdamSourceClass* par_source_class = (GdamSourceClass*)parent_class;
	int type;
	if (gdam_buffer_peek_int(command, &type)) {
		if (type == GDAM_CFG_MIX_ADD) {
			GdamMix* mix;
			GdamSource* subsource;
			int source_id;
			GdamTime when;
			mix = GDAM_MIX(source);
			gdam_buffer_parse_int(command, &type);
			if (!gdam_buffer_parse_int(command, &source_id)) {
			 	*was_ok = FALSE;
				gdam_debug("mix_add: no source_id ");
				return NULL;
			}
			subsource = gdam_namespace_local_source(source_id);
			if (!subsource) {
			 	*was_ok = FALSE;
				gdam_debug("mix_add: bad source id %d",
					source_id);
				return NULL;
			}

			if (!gdam_buffer_parse_time(command, &when)) {
				*was_ok = FALSE;
				gdam_debug("mix_add: no time (too short)");
				return NULL;
			}
			gdam_dispatch_log("add %d %d %lld",
					source->source_id,
					source_id,
					when);
			gdam_mix_connect(mix, subsource, when);
			*was_ok = TRUE;
			return NULL;
		}
	}
	return par_source_class->interpret_command(source, command, was_ok);
}

int gdam_mix_num_sources(GdamMix* mix)
{
	GdamMixElement* element;
	int count = 0;
	for (element = mix->first_source;
	     element;
	     element = element->next)
	     	count++;
	return count;
}
static void gdam_mix_get_subsources   (GdamSource*      source,
                                       int*             num_subs,
				       GdamSource***    subs)
{
	GdamMix* mix = GDAM_MIX(source);
	int count = 0;
	GdamMixElement* element;
	g_return_if_fail(mix != NULL);
	*num_subs = gdam_mix_num_sources(mix);
	*subs = g_new(GdamSource*, *num_subs);
	for (element = mix->first_source;
	     element;
	     element = element->next)
	     	(*subs)[count++] = element->sub_source;
}

GtkType gdam_mix_get_type()
{
	static GtkType mix_type = 0;
	if (!mix_type) {
		static const GtkTypeInfo mix_info = {
			"GdamMix",
			sizeof(GdamMix),
			sizeof(GdamMixClass),
			(GtkClassInitFunc) gdam_mix_class_init,
			(GtkObjectInitFunc) gdam_mix_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		parent_class = gtk_type_class(GDAM_TYPE_SOURCE);
		mix_type = gtk_type_unique(GDAM_TYPE_SOURCE, &mix_info);
	}
	return mix_type;
}
