/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdambender.h"

#define MIN_BUFFERING	(44100 * 2)

/* XXX: we could probably bear much larger numbers,
 *      but they really are of questionable value... */
#define MIN_RATE	(-1000 * GdamFixed_ONE)
#define MAX_RATE	(+1000 * GdamFixed_ONE)

static void        gdam_bender_miss_data        (GdamSource*      input,
					         int              lost);
static gboolean    gdam_bender_is_ready_to_read (GdamSource*      source);
static int	   gdam_bender_read             (GdamSource*      input,
					         GdamSample*      samples,
					         int              max_read);
static GdamTime    gdam_bender_get_position     (GdamSource* source);
static void        gdam_bender_get_latency_info (GdamSource*      source,
                                                 GdamSource*      subsource,
                                                 GdamLatencyInfo* latency);

static void remove_change(GdamBender* bender);
static void remove_new_position_changes(GdamBender* bender);
static void gdam_bender_init(GdamBender* bender)
{
	bender->rate = GdamFixed_ONE;
	bender->fade_window = 64;
}
static void append_changes_to_list(GdamBenderChange** p_last, 
                                   int                num_changes,
				   int*               time_rates)
{
	int i;
	for (i = 0; i < num_changes; i++) {
		GdamBenderChange* change;

		if (time_rates[i*2] <= 0)
			continue;
		
		change = g_new(GdamBenderChange, 1);
		*p_last = change;
		p_last = &change->next;
		change->next = NULL;
		change->num_samples = time_rates[i*2+0];
		change->new_rate = CLAMP(time_rates[i*2+1], MIN_RATE, MAX_RATE);
	}
}
static void gdam_bender_arg_set_changes(GtkObject* object, GdamArg* arg)
{
	GdamBender* bender = GDAM_BENDER(object);
	int num_changes = GDAM_ARG_INT_ARRAY(arg).length / 2;
	int* time_rates = GDAM_ARG_INT_ARRAY(arg).array;
	while (bender->changes)
		remove_change(bender);
	append_changes_to_list(&bender->changes, num_changes, time_rates);
	if (bender->change_mode != GDAM_BENDER_CHANGE_KEEP_TIME
	 && bender->change_mode != GDAM_BENDER_CHANGE_SNAP_TIME)
		bender->end_offset = bender->backbuffer_offset;
	remove_new_position_changes(bender);
}
static void gdam_bender_arg_append_changes(GtkObject* object, GdamArg* arg)
{
	int num_changes = GDAM_ARG_INT_ARRAY(arg).length / 2;
	int* time_rates = GDAM_ARG_INT_ARRAY(arg).array;
	GdamBender* bender = GDAM_BENDER(object);
	GdamBenderChange** p_last = &bender->changes;
	gboolean was_empty = (bender->changes == NULL);
	while (*p_last != NULL)
		p_last = &((*p_last)->next);
	append_changes_to_list(&bender->changes, num_changes, time_rates);
	if (was_empty) {
		if (bender->change_mode != GDAM_BENDER_CHANGE_KEEP_TIME
		 && bender->change_mode != GDAM_BENDER_CHANGE_SNAP_TIME)
			bender->end_offset = bender->backbuffer_offset;
		remove_new_position_changes(bender);
	}
}
static void gdam_bender_arg_get_changes(GtkObject* object, GdamArg* arg)
{
	int i, num_changes = 0;
	GdamBenderChange* change;
	GdamBender* bender = GDAM_BENDER(object);
	int* array;
	for (change = bender->changes; change != NULL; change = change->next) {
		num_changes++;
	}
	GDAM_ARG_INT_ARRAY(arg).length = num_changes * 2;
	GDAM_ARG_INT_ARRAY(arg).array = array = g_new(int, num_changes * 2);
	i = 0;
	for (change = bender->changes; change != NULL; change = change->next) {
		array[i++] = change->num_samples;
		array[i++] = change->new_rate;
	}
}
static void gdam_bender_arg_set_change_mode(GtkObject* object, GdamArg* arg)
{
	const char* string = GDAM_ARG_STRING(arg);
	GdamBender* bender = GDAM_BENDER(object);
	if (strcmp(string, "continuous") == 0) {
		bender->change_mode = GDAM_BENDER_CHANGE_CONTINUOUS;
	} else if (strcmp(string, "keep_time") == 0) {
		bender->change_mode = GDAM_BENDER_CHANGE_KEEP_TIME;
		bender->end_offset = bender->backbuffer_offset;
	} else if (strcmp(string, "snap_time") == 0) {
		bender->change_mode = GDAM_BENDER_CHANGE_SNAP_TIME;
		bender->end_offset = bender->backbuffer_offset;
	} else if (strcmp(string, "return") == 0) {
		bender->change_mode = GDAM_BENDER_CHANGE_RETURN;
	} else {
		gdam_debug("gdam_bender no such change_mode: %s", string);
	}
}

static void gdam_bender_arg_get_change_mode(GtkObject* object, GdamArg* arg)
{
	const char* string = NULL;
	GdamBender* bender = GDAM_BENDER(object);
	switch (bender->change_mode) {
		case GDAM_BENDER_CHANGE_CONTINUOUS:
			string = "continuous";
			break;
		case GDAM_BENDER_CHANGE_KEEP_TIME:
			string = "keep_time";
			break;
		case GDAM_BENDER_CHANGE_SNAP_TIME:
			string = "snap_time";
			break;
		case GDAM_BENDER_CHANGE_RETURN:
			string = "return";
			break;
		default:
			g_warning("change_mode was invalid in bender");
			return;
	}
	GDAM_ARG_STRING(arg) = g_strdup(string);
}
static void gdam_bender_arg_jump_to_mark(GtkObject* object, GdamArg* arg)
{
	GdamBender* bender = GDAM_BENDER(object);
	gint64 offset = (gint64)GDAM_ARG_INT(arg) * GdamFixed_ONE;
	bender->backbuffer_offset = bender->end_offset + offset;
}
static void gdam_bender_arg_set_mark(GtkObject* object, GdamArg* arg)
{
	GdamBender* bender = GDAM_BENDER(object);
	gint64 offset = (gint64)GDAM_ARG_INT(arg) * GdamFixed_ONE;
	bender->end_offset = bender->backbuffer_offset + offset;
}
static void gdam_bender_arg_set_rate(GtkObject* object, GdamArg* arg)
{
	GDAM_BENDER(object)->rate = GDAM_ARG_INT(arg);
}
static void gdam_bender_arg_get_rate(GtkObject* object, GdamArg* arg)
{
	GDAM_ARG_INT(arg) = GDAM_BENDER(object)->rate;
}
static void gdam_bender_arg_set_readahead(GtkObject* object, GdamArg* arg)
{
	GDAM_BENDER(object)->readahead = GDAM_ARG_INT(arg);
}
static void gdam_bender_arg_get_readahead(GtkObject* object, GdamArg* arg)
{
	GDAM_ARG_INT(arg) = GDAM_BENDER(object)->readahead;
}
static void gdam_bender_arg_set_buffer_length(GtkObject* object, GdamArg* arg)
{
	gdam_buffering_filter_ensure_space(
			GDAM_BUFFERING_FILTER(object),
			GDAM_ARG_INT(arg)
	);
}
static void gdam_bender_arg_get_buffer_length(GtkObject* object, GdamArg* arg)
{
	GDAM_ARG_INT(arg) = GDAM_BUFFERING_FILTER(object)->buffer_length;
}

static void gdam_bender_arg_set_fade_window(GtkObject* object, GdamArg* arg)
{
	GdamBender* bender = GDAM_BENDER(object);
	int new_window = GDAM_ARG_INT(arg);
	if (bender->fade_window == 0)
		bender->fade_remaining = 0;
	else
		bender->fade_remaining = bender->fade_remaining
		                       * new_window
				       / bender->fade_window;
	bender->fade_window = new_window;
}

static void gdam_bender_arg_get_fade_window(GtkObject* object, GdamArg* arg)
{
	GdamBender* bender = GDAM_BENDER(object);
	GDAM_ARG_INT(arg) = bender->fade_window;
}

#if 0
static void set_relative_smoothed_position(GdamBender* bender, int relpos)
{
	int projected;
	if (bender->last_smoothed_length <= 2)
		projected = 4096;
	else
		projected = bender->last_smoothed_length;
	bender->last_smoothed_length = 0;

	/*
	 * We model the situation by the differential equation:
	 *
	 *       v'(t) = (-drag) * v(t) + k
	 *
	 * where `k' is the accelation we are going to apply.
	 *
	 * It has the following boundary conditions,
	 *
	 *	v(0)       = v0
	 *	v(endtime) = endpos
	 *
	 */
	/* The solution for the first-order DE is taken from a book... */
	c2 = (gdouble) relpos + rate * projected
	   / (1.0 + drag * relpos - exp(-drag * relpos));
	k = c2 * drag * drag - rate * drag;

	/* Then position and velocity are given:
	 *
	 *	p(t) = -c2 * exp(-drag*t) + k*t/drag + c2
	 *	v(t) = -c2 * drag * exp(-drag*t) + k*t/drag + c2
	 */

	...

}

static void gdam_bender_arg_set_smoothed_position(GtkObject* object,
                                                  GdamArg* arg)
{
	GdamBender* bender = GDAM_BENDER(object);
	gint64 position = GDAM_ARG_INT64(arg);
	GdamSource* filtered = GDAM_FILTER(object)->filtered;
	gint64 our_pos;
	int rel_pos;
	if (filtered == NULL)
		return;
	our_pos = gdam_source_get_position(filtered) - bender->readahead;
	rel_pos = (int)(our_pos - position);

	set_relative_smoothed_position(bender, rel_pos);
}
#endif

static void gdam_bender_class_init(GdamBenderClass* bender_class)
{
	GdamSourceClass* source_class = GDAM_SOURCE_CLASS(bender_class);
	//if (!parent_class)
		//parent_class = gtk_type_class(GDAM_TYPE_BUFFERING_FILTER);
	source_class->read = gdam_bender_read;
	source_class->miss_data = gdam_bender_miss_data;
	source_class->is_ready_to_read = gdam_bender_is_ready_to_read;
	source_class->get_latency_info = gdam_bender_get_latency_info;
	source_class->get_position = gdam_bender_get_position;
	gdam_arg_register(GTK_OBJECT_CLASS(bender_class),
	                  "changes",
			  GDAM_ARG_TYPE_INT_ARRAY,
			  gdam_bender_arg_set_changes,
			  gdam_bender_arg_get_changes,
			  GDAM_ARG_VITAL);
	gdam_arg_register(GTK_OBJECT_CLASS(bender_class),
	                  "append_changes",
			  GDAM_ARG_TYPE_INT_ARRAY,
			  gdam_bender_arg_append_changes,
			  NULL,
			  0);
	gdam_arg_register(GTK_OBJECT_CLASS(bender_class),
	                  "change_mode",
			  GDAM_ARG_TYPE_STRING,
			  gdam_bender_arg_set_change_mode,
			  gdam_bender_arg_get_change_mode,
			  GDAM_ARG_VITAL);
	gdam_arg_register(GTK_OBJECT_CLASS(bender_class),
	                  "rate",
			  GDAM_ARG_TYPE_INT,
			  gdam_bender_arg_set_rate,
			  gdam_bender_arg_get_rate,
			  GDAM_ARG_VITAL);
	gdam_arg_register(GTK_OBJECT_CLASS(bender_class),
	                  "readahead",
			  GDAM_ARG_TYPE_INT,
			  gdam_bender_arg_set_readahead,
			  gdam_bender_arg_get_readahead,
			  GDAM_ARG_VITAL);
	gdam_arg_register(GTK_OBJECT_CLASS(bender_class),
	                  "jump_to_mark",
			  GDAM_ARG_TYPE_INT,
			  gdam_bender_arg_jump_to_mark,
			  NULL,
			  0);
	gdam_arg_register(GTK_OBJECT_CLASS(bender_class),
	                  "set_mark",
			  GDAM_ARG_TYPE_INT,
			  gdam_bender_arg_set_mark,
			  NULL,
			  0);
	gdam_arg_register(GTK_OBJECT_CLASS(bender_class),
	                  "buffer_length",
			  GDAM_ARG_TYPE_INT,
			  gdam_bender_arg_set_buffer_length,
			  gdam_bender_arg_get_buffer_length,
			  GDAM_ARG_VITAL);
	gdam_arg_register(GTK_OBJECT_CLASS(bender_class),
	                  "fade_window",
			  GDAM_ARG_TYPE_INT,
			  gdam_bender_arg_set_fade_window,
			  gdam_bender_arg_get_fade_window,
			  GDAM_ARG_VITAL);
}

GtkType gdam_bender_get_type()
{
	static GtkType bender_type = 0;
	if (!bender_type) {
		static const GtkTypeInfo bender_info = {
			"GdamBender",
			sizeof(GdamBender),
			sizeof(GdamBenderClass),
			(GtkClassInitFunc) gdam_bender_class_init,
			(GtkObjectInitFunc) gdam_bender_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GDAM_TYPE_BUFFERING_FILTER;
		bender_type = gtk_type_unique(parent, &bender_info);
	}
	return bender_type;
}

static void gdam_bender_miss_data(GdamSource*      input,
				  int              lost)
{
	GdamSource* filtered;
	filtered = GDAM_FILTER(input)->filtered;
	if (filtered != NULL)
		gdam_source_miss_data(filtered, lost);
	else
		GDAM_BENDER(input)->backbuffer_offset += ((gint64)lost << 16);
}

static void gdam_bender_get_latency_info(GdamSource* source,
                                         GdamSource* subsource,
                                         GdamLatencyInfo* latency)
{
	GdamBender* bender = GDAM_BENDER(source);
	int bb_offset = (int)(bender->backbuffer_offset >> 16);
	int bb_len = GDAM_BUFFERING_FILTER(source)->buffer_current_amount;
	latency->latency = bb_len - bb_offset;
	latency->resample = bender->rate;
}

static gboolean gdam_bender_is_ready_to_read(GdamSource*      source)
{
	GdamBender* bender;
	GdamFixed rate;
	GdamSource* filtered;
	bender = GDAM_BENDER(source);
	if (bender->changes != NULL)
		rate = bender->changes->new_rate;
	else
		rate = bender->rate;
	if (rate <= 0)
		return TRUE;
	filtered = GDAM_FILTER(source)->filtered;
	if (filtered == NULL)
		return TRUE;
	return gdam_source_is_ready_to_read(filtered);
}

static int do_filtered_read(GdamBender* bender, int num)
{
	GdamSource* filtered = ((GdamFilter*)bender)->filtered;
	int num_read, discard;
	GdamBufferingFilter* buffering_filter = (GdamBufferingFilter*)bender;
	GdamSample* buf;
	buf = alloca(sizeof(GdamSample) * num);
	num_read = gdam_source_read(filtered, buf, num);
	if (num_read <= 0)
		return 0;
	discard = gdam_buffering_filter_append(buffering_filter, buf, num_read);
	if (discard) {
		bender->backbuffer_offset -= (gint64) discard * GdamFixed_ONE;
		bender->end_offset -= (gint64) discard * GdamFixed_ONE;
	}
	return num_read;
}

static void remove_change(GdamBender* bender)
{
	GdamBenderChange* next;
	g_return_if_fail(bender->changes != NULL);
	next = bender->changes->next;
	g_free(bender->changes);
	bender->changes = next;
	if (next == NULL
	 && bender->change_mode == GDAM_BENDER_CHANGE_SNAP_TIME) {
		bender->backbuffer_offset = bender->end_offset;
		bender->fade_remaining = bender->fade_window;
	}
}
static void remove_new_position_changes(GdamBender* bender)
{
	while (bender->changes
	    && bender->changes->num_samples == 0) {
	    	bender->backbuffer_offset += (gint64)bender->changes->new_rate
		                                 *   GdamFixed_ONE;
		remove_change(bender);
		bender->fade_remaining = bender->fade_window;
	}
}

static void do_fading(GdamBender* bender,
                      GdamSample* sample,
		      int         num)
{
	int to_fade;
	int fade_remaining;
	GdamSample fade_sample = bender->fade_sample;
	fade_remaining = bender->fade_remaining;
	to_fade = num;
	if (to_fade > fade_remaining)
		to_fade = fade_remaining;
	while (to_fade-- > 0) {
		GdamSampleProduct tmp;

		tmp = (GdamSampleProduct)fade_remaining * fade_sample.s[0]
		    + (GdamFixed_ONE - fade_remaining)  * sample->s[0];
		tmp >>= GdamFixed_LOG_ONE;
		sample->s[0] = (GdamSampleValue) tmp;

		tmp = (GdamSampleProduct)fade_remaining * fade_sample.s[1]
		    + (GdamFixed_ONE - fade_remaining)  * sample->s[1];
		tmp >>= GdamFixed_LOG_ONE;
		sample->s[1] = (GdamSampleValue) tmp;

		sample++;
		fade_remaining--;
	}
	bender->fade_remaining = fade_remaining;
}

/*=====================================================================*
 *                             compute_clip                            *
 *                                                                     *
 * Determine how many samples from a range need to be zeroed out,      *
 * in the front and the back because they are outside the buffer.      *
 *                                                                     *
 * Parameters:                                                         *
 *     bender            the bender to clip for.                       *
 *     rate              the forward or backward velocity of the       *
 *                       position.                                     *
 *     offset            the position within the backbuffer.           *
 *     num               the number of samples of output at that rate. *
 *     prepad_out        returns the number of samples of silence to   *
 *                       write before sound starts.                    *
 *     postpad_out       returns the number of samples of silence to   *
 *                       write after sound ends.                       *
 *     return            whether any part was in the backbuffer.       *
 *=====================================================================*/
static gboolean
compute_clip(GdamBender*   bender,
             GdamFixed     rate,
	     gint64        offset,
	     int           num,
	     int*          prepad_out,
	     int*          postpad_out)
{
	gint64 end_offset;
	gint64 current_amount;
	gboolean starts_in_bounds, ends_in_bounds;
	gint rate_minus_one = rate - 1;
	current_amount = GDAM_BUFFERING_FILTER(bender)->buffer_current_amount;
	current_amount *= GdamFixed_ONE;
	end_offset = offset + (gint64)rate * num;

	*prepad_out = 0;
	*postpad_out = 0;

	starts_in_bounds = (0 <= offset && offset < current_amount);
	ends_in_bounds = (0 <= end_offset && end_offset < current_amount);

	g_return_val_if_fail(rate != 0, FALSE);

	if (!starts_in_bounds) {
		gint64 in_point = rate > 0 ? 0 : current_amount - 1;
		*prepad_out = (in_point - offset + rate_minus_one) / rate;
		if (*prepad_out < 0)
			return FALSE;
	}
	if (!ends_in_bounds) {
		gint64 in_point = rate > 0 ? current_amount - 1 : 0;
		*postpad_out = (end_offset - in_point + rate_minus_one) / rate;
		if (*postpad_out < 0)
			return FALSE;
	}
	if (*prepad_out + *postpad_out >= num)
		return FALSE;
	return TRUE;
}

/*=====================================================================*
 *                      resample_from_back_buffer                      *
 *                                                                     *
 * Fills a buffer with sound (possibly padded with silence at          *
 * the ends) from the back buffer.  It resamples forward               *
 * or backward.                                                        *
 *                                                                     *
 * Parameters:                                                         *
 *     bender            the bender to read data from.                 *
 *     rate              the resampling rate (may be negative).        *
 *     offset            the backbuffer offset (with fraction).        *
 *     num_from_back     the number of samples to read from the        *
 *                       backbuffer.                                   *
 *     output            the output buffer.                            *
 *=====================================================================*/
static void resample_from_back_buffer(GdamBender* bender,
                                      GdamFixed   rate,
				      gint64      offset,
				      int         num_from_back,
				      GdamSample* output)
{
	gint64 max_offset;
	int num_left;
	int prepad, postpad;
	GdamBufferingFilter* buf_filter = (GdamBufferingFilter*)bender;
	gboolean had_new_sample = FALSE;

	num_left = num_from_back;

	if (rate == 0) {
		gdam_sample_zero(output, num_from_back);
		return;
	}

	/* Now clip.  Set prepad and postpad to the number of
	 * zeroes where data is unavailable,
	 * and adjust num_left/output to be the number of samples
	 * to write from the real backbuffer. */
	if (!compute_clip(bender, rate, offset, num_left, &prepad, &postpad)) {
		gdam_sample_zero(output, num_from_back);
		return;
	}

	gdam_sample_zero(output, prepad);
	output += prepad;
	num_left -= (postpad + prepad);
	offset += (gint64) rate * prepad;
	max_offset = (gint64) buf_filter->buffer_current_amount * GdamFixed_ONE;

	g_return_if_fail(offset >= 0);
	g_return_if_fail(offset <= max_offset);
	while (num_left-- > 0) {
		*output++ = gdam_buffering_filter_sample_from_start
						(buf_filter,
						 (int)(offset >> 16));
		offset += rate;
		had_new_sample = TRUE;
	}
	if (offset > max_offset) {
		g_message("offset=%lld, max_offset=%lld, rate=%d, num=%d",
		offset,max_offset, rate,num_from_back);
	}
	g_return_if_fail(offset >= 0);
	g_return_if_fail(offset <= max_offset);

}

static GdamTime gdam_bender_get_position(GdamSource* source)
{
	GdamSource* filtered;
	GdamBufferingFilter* buf_filter;
	GdamTime rv;
	GdamBender* bender;
	buf_filter = GDAM_BUFFERING_FILTER(source);
	filtered = GDAM_FILTER(source)->filtered;
	bender = GDAM_BENDER(source);
	if (filtered == NULL)
		return -1;
	rv = gdam_source_get_position(filtered);
	{
		int latency;
		latency = (bender->backbuffer_offset>>16)
		        - (buf_filter->buffer_current_amount);
		if (bender->rate != 0)
			latency = (gint64)latency * GdamFixed_ONE
			         / bender->rate;
		rv -= latency;
	}
	return rv;
}

static int  gdam_bender_read(GdamSource*      source,
			     GdamSample*      samples,
			     int              max_read)
{
	GdamBender* bender;
	GdamSource* filtered;
	GdamBufferingFilter* buffering;
	int rv = 0;

	bender = GDAM_BENDER(source);
	filtered = GDAM_FILTER(source)->filtered;
	buffering = GDAM_BUFFERING_FILTER(source);
	if (filtered == NULL)
		return 0;

	/* Make sure we have enough space in the buffer. */
	if (buffering->buffer_length < MIN_BUFFERING)
		gdam_buffering_filter_ensure_space(buffering, MIN_BUFFERING);

	if (bender->change_mode == GDAM_BENDER_CHANGE_KEEP_TIME
	 || bender->change_mode == GDAM_BENDER_CHANGE_SNAP_TIME
	 || bender->readahead > 0) {
		int advance;
		gint64 end_time;
		end_time = bender->backbuffer_offset + bender->readahead;
		if (bender->change_mode == GDAM_BENDER_CHANGE_KEEP_TIME
		 || bender->change_mode == GDAM_BENDER_CHANGE_SNAP_TIME) {
			gint64 est_end;
			est_end = (gint64)max_read * bender->rate
			        + bender->end_offset;
			if (end_time < est_end) {
				end_time = est_end;
			}
		}
		end_time -= (gint64)buffering->buffer_current_amount
		                            * GdamFixed_ONE;
		advance = (int)(end_time / GdamFixed_ONE);
		if (advance > 0 && gdam_source_is_ready_to_read(filtered))
			(void) do_filtered_read(bender, advance);
	}
	
	while (rv < max_read) {
		int max_time;
		GdamFixed rate;
		gboolean had_new_sample = FALSE;

		/* Compute `rate' and `max_time'. */
		max_time = max_read - rv;
		while (bender->changes
		    && bender->changes->num_samples <= 0) {
			remove_change(bender);
			remove_new_position_changes(bender);
		}
		if (bender->changes != NULL) {
			rate = bender->changes->new_rate;
			if (max_time > bender->changes->num_samples)
				max_time = bender->changes->num_samples;
		} else {
			rate = bender->rate;
		}

		/*
		 * If we are resampling forward, sometimes we 
		 * should try to read from the source to keep up.
		 */
		if (rate > 0) {
			gint64 end_position;
			int current_length = buffering->buffer_current_amount;
			end_position = bender->backbuffer_offset
			             + (gint64)rate * (max_time+1);
			end_position >>= GdamFixed_LOG_ONE;
		   	if (end_position > current_length) {
				int to_read;
				to_read = end_position + 1 - current_length;
				if (!gdam_source_is_ready_to_read(filtered)) {
					/* XXX: this can really happen!!! */
					g_return_val_if_fail(rv > 0, 1);
					break;
				}
				if (!do_filtered_read(bender, to_read))
					break;
				had_new_sample = TRUE;
			}
		}

		resample_from_back_buffer(bender,
		                          rate,
					  bender->backbuffer_offset,
					  max_time,
					  samples + rv);
		rv += max_time;

		if (bender->changes != NULL) {
			bender->changes->num_samples -= max_time;
			if (bender->changes->num_samples <= 0) {
				remove_change(bender);
				remove_new_position_changes(bender);
			}
		}
		bender->backbuffer_offset += max_time * rate;

		if (had_new_sample) {
			bender->last_sample
				= gdam_buffering_filter_sample_at(buffering, 0);
		}
		do_fading(bender, samples + rv, max_time);

	}
	bender->end_offset += (gint64)bender->rate * rv;
	return rv;
}
