/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gtkpixbutton.h"
#include "pixmapcache.h"
#include "gdamskin.h"		/* needed for path stuff. */
#include <gtk/gtksignal.h>
#include <gtk/gtkmain.h>

enum {
	PRESSED,
	RELEASED,
	CLICKED,
	LAST_SIGNAL
};
static guint button_signals[LAST_SIGNAL] = { 0 };

static int pix_button_expose_event          (GtkWidget*       widget,
                                             GdkEventExpose*  event);
static int pix_button_configure_event       (GtkWidget*       widget,
                                             GdkEventConfigure*event);
static void pix_button_draw                 (GtkWidget*       widget,
                                             GdkRectangle*    area);
static void gtk_pix_button_draw_default (GtkWidget *widget);
static void gtk_pix_button_draw_focus (GtkWidget *widget);
static void pix_button_size_request         (GtkWidget*       widget,
                                             GtkRequisition*  event);
static int pix_button_button_press_event    (GtkWidget*       widget,
                                             GdkEventButton*  event);
static int pix_button_button_release_event  (GtkWidget*       widget,
                                             GdkEventButton*  event);
static int pix_button_button_release_event  (GtkWidget*       widget,
                                             GdkEventButton*  event);
static int pix_button_enter_notify_event    (GtkWidget*       widget,
                                             GdkEventCrossing*  event);
static int pix_button_leave_notify_event    (GtkWidget*       widget,
                                             GdkEventCrossing*  event);
static GtkObjectClass* parent_class;
static void load_images                     (GtkPixButton* pix_button);
static void gtk_pix_button_init(GtkPixButton* pix_button)
{
	
	GTK_WIDGET_SET_FLAGS(pix_button, GTK_CAN_FOCUS|GTK_RECEIVES_DEFAULT);
	gtk_widget_add_events(GTK_WIDGET(pix_button),
	                        GDK_EXPOSURE_MASK
			      | GDK_STRUCTURE_MASK
			      | GDK_ENTER_NOTIFY_MASK
			      | GDK_LEAVE_NOTIFY_MASK);
	pix_button->state = GTK_STATE_NORMAL;
}
static void gtk_pix_button_class_init(GtkPixButtonClass* klass)
{
	GtkWidgetClass* widget_class = GTK_WIDGET_CLASS(klass);
	GtkObjectClass* object_class = GTK_OBJECT_CLASS(klass);
	widget_class->expose_event = pix_button_expose_event;
	widget_class->configure_event = pix_button_configure_event;
	widget_class->button_press_event = pix_button_button_press_event;
	widget_class->button_release_event = pix_button_button_release_event;
	widget_class->enter_notify_event = pix_button_enter_notify_event;
	widget_class->leave_notify_event = pix_button_leave_notify_event;
	widget_class->size_request = pix_button_size_request;
	widget_class->draw = pix_button_draw;
	widget_class->draw_default = gtk_pix_button_draw_default;
	widget_class->draw_focus = gtk_pix_button_draw_focus;
	button_signals[PRESSED] =
		gtk_signal_new ("pressed",
				GTK_RUN_FIRST,
				object_class->type,
				GTK_SIGNAL_OFFSET(GtkPixButtonClass, pressed),
				gtk_marshal_NONE__NONE,
				GTK_TYPE_NONE, 0);
	button_signals[RELEASED] =
		gtk_signal_new ("released",
				GTK_RUN_FIRST,
				object_class->type,
				GTK_SIGNAL_OFFSET(GtkPixButtonClass, released),
				gtk_marshal_NONE__NONE,
				GTK_TYPE_NONE, 0);
	button_signals[CLICKED] =
		gtk_signal_new ("clicked",
				GTK_RUN_FIRST | GTK_RUN_ACTION,
				object_class->type,
				GTK_SIGNAL_OFFSET(GtkPixButtonClass, clicked),
				gtk_marshal_NONE__NONE,
				GTK_TYPE_NONE, 0);
	gtk_object_class_add_signals(object_class,
	                             button_signals,
				     LAST_SIGNAL);
}

GtkType gtk_pix_button_get_type()
{
	static GtkType pix_button_type = 0;
	if (!pix_button_type) {
		static const GtkTypeInfo pix_button_info = {
			"GtkPixButton",
			sizeof(GtkPixButton),
			sizeof(GtkPixButtonClass),
			(GtkClassInitFunc) gtk_pix_button_class_init,
			(GtkObjectInitFunc) gtk_pix_button_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GTK_TYPE_EVENT_BOX;
		parent_class = gtk_type_class(parent);
		pix_button_type = gtk_type_unique(parent,
		                                 &pix_button_info);
	}
	return pix_button_type;
}

static void pix_button_draw(GtkWidget* widget, GdkRectangle* area)
{
	GtkPixButton* button = GTK_PIX_BUTTON(widget);
	GdkWindow* window = widget->window;
	GdkPixmap* pixmap;
	if (!GTK_WIDGET_DRAWABLE(widget)) {
		return;
	}
	if (!button->loaded_images)
		load_images(button);
	pixmap = button->pixmaps[button->state];

	gdk_window_set_back_pixmap (widget->window, NULL, TRUE);
	gdk_window_clear_area (widget->window,
	                       area->x, area->y,
			       area->width, area->height);
	if (GTK_WIDGET_HAS_DEFAULT (widget) &&
	   GTK_BUTTON (widget)->relief == GTK_RELIEF_NORMAL) {
		gtk_paint_box (widget->style, widget->window,
				GTK_STATE_NORMAL, GTK_SHADOW_IN,
				area, widget, "buttondefault",
				area->x, area->y, area->width, area->height);
	}


	gdk_gc_set_clip_mask(widget->style->fg_gc[0],
	                     button->masks[button->state]);
	                     
	gdk_window_copy_area(window,
	                     widget->style->fg_gc[0],
			     0,
			     0,
			     pixmap,
			     0,
			     0,
			     widget->allocation.width,
			     widget->allocation.height);

	gdk_gc_set_clip_mask(widget->style->fg_gc[0],
	                     NULL);
}
static void
gtk_pix_button_draw_focus (GtkWidget *widget)
{
  gtk_widget_draw (widget, NULL);
}

static void
gtk_pix_button_draw_default (GtkWidget *widget)
{
  gtk_widget_draw (widget, NULL);
}

static int pix_button_button_press_event(GtkWidget*       widget,
                                         GdkEventButton*  event)
{
	GtkPixButton* button = GTK_PIX_BUTTON(widget);
	if (button->state != GTK_STATE_PRELIGHT
	 && button->state != GTK_STATE_NORMAL) {
		return TRUE;
	}
	button->state = GTK_STATE_SELECTED;
	gtk_pix_button_pressed(button);
	gtk_widget_draw_default(GTK_WIDGET(button));
	return TRUE;
}

static int pix_button_button_release_event(GtkWidget*       widget,
                                           GdkEventButton*  event)
{
	gboolean clicked = FALSE;
	GtkPixButton* button = GTK_PIX_BUTTON(widget);
	if (button->state != GTK_STATE_SELECTED
	 && button->state != GTK_STATE_ACTIVE)
	 	return TRUE;
	
	if (button->state == GTK_STATE_SELECTED) {
		button->state = GTK_STATE_PRELIGHT;
		clicked = TRUE;
	} else
		button->state = GTK_STATE_NORMAL;
	gtk_pix_button_released(button);
	if (clicked)
		gtk_pix_button_clicked(button);
	gtk_widget_draw_default(GTK_WIDGET(button));
	return TRUE;
}
static void load_images(GtkPixButton* pix_button)
{
	GdkWindow* window = GTK_WIDGET(pix_button)->window;
	int i;
	if (pix_button->loaded_images)
		return;
	pix_button->loaded_images = TRUE;
	for (i = 0; i < 4; i++) {
		PixmapInfo* info;
		char* xpm = pix_button->xpm_files[i];
		const char* fname;
		if (xpm == NULL) {
			g_warning("error: no file %d for pixbutton", i);
			continue;
		}
		fname = gdam_skin_search_for_file(xpm);
		if (fname == NULL) {
			g_warning("error: no image %d, %s for pixbutton", i,
				xpm);
			continue;
		}
		info = pixmap_cache_get_pixmap(window, fname);
		if (info == NULL) {
			g_warning("error loading pixmap %s", fname);
			return;
		}
		pix_button->pixmaps[i] = info->pixmap;
		pix_button->masks[i] = info->mask;
		pix_button->sizes[i].x = info->width;
		pix_button->sizes[i].y = info->height;
	}
}

static int pix_button_expose_event(GtkWidget*       widget,
                                   GdkEventExpose*  event)
{
	GtkPixButton* button = GTK_PIX_BUTTON(widget);
	gtk_widget_draw_default(GTK_WIDGET(button));
	return FALSE;
}
static int pix_button_configure_event       (GtkWidget*       widget,
                                             GdkEventConfigure*event)
{
	GtkPixButton* button = GTK_PIX_BUTTON(widget);
	if (!button->loaded_images)
		load_images(button);
	gtk_widget_draw_default(GTK_WIDGET(button));
	return FALSE;
}
static int pix_button_enter_notify_event(GtkWidget*       widget,
                                         GdkEventCrossing*event)
{
	GtkPixButton* button = GTK_PIX_BUTTON(widget);
	GtkWidget* event_widget;
	event_widget = gtk_get_event_widget((GdkEvent*)event);
	if (event_widget != widget
	 || event->detail == GDK_NOTIFY_INFERIOR)
		return FALSE;
	if (button->state != GTK_STATE_NORMAL
	 && button->state != GTK_STATE_ACTIVE)
	 	return FALSE;
	if (button->state == GTK_STATE_NORMAL)
		button->state = GTK_STATE_PRELIGHT;
	else
		button->state = GTK_STATE_SELECTED;
	gtk_widget_draw_default(widget);
	return FALSE;
}
static int pix_button_leave_notify_event(GtkWidget*       widget,
                                         GdkEventCrossing*event)
{
	GtkPixButton* button = GTK_PIX_BUTTON(widget);
	GtkWidget* event_widget;
	event_widget = gtk_get_event_widget((GdkEvent*)event);
	if (event_widget != widget
	 || event->detail == GDK_NOTIFY_INFERIOR)
		return FALSE;
	if (button->state != GTK_STATE_PRELIGHT
	 && button->state != GTK_STATE_SELECTED)
	 	return FALSE;
	if (button->state == GTK_STATE_PRELIGHT)
		button->state = GTK_STATE_NORMAL;
	else
		button->state = GTK_STATE_ACTIVE;
	gtk_widget_draw_default(widget);
	return FALSE;
}
static void pix_button_size_request    (GtkWidget*       widget,
                                        GtkRequisition*  requisition)
{
	GtkPixButton* pix_button = GTK_PIX_BUTTON(widget);
	int state = pix_button->state;
	requisition->width = pix_button->sizes[state].x;
	requisition->height = pix_button->sizes[state].y;
}

void gtk_pix_button_pressed(GtkPixButton* pix_button)
{
	gtk_signal_emit(GTK_OBJECT(pix_button),
	                button_signals[PRESSED]);
}

void gtk_pix_button_released(GtkPixButton* pix_button)
{
	gtk_signal_emit(GTK_OBJECT(pix_button),
	                button_signals[RELEASED]);
}
void gtk_pix_button_clicked(GtkPixButton* pix_button)
{
	gtk_signal_emit(GTK_OBJECT(pix_button),
	                button_signals[CLICKED]);
}

void gtk_pix_button_set_pixmaps_by_filename(GtkPixButton* button,
                                            const char*   normal,
                                            const char*   active,
					    const char*   prelight,
					    const char*   selected)
{
	button->xpm_files[GTK_STATE_NORMAL] = g_strdup(normal);
	button->xpm_files[GTK_STATE_ACTIVE] = g_strdup(active);
	button->xpm_files[GTK_STATE_PRELIGHT] = g_strdup(prelight);
	button->xpm_files[GTK_STATE_SELECTED] = g_strdup(selected);
	if (GTK_WIDGET(button)->window)
		load_images(button);
}

GtkWidget* gtk_pix_button_create_func      (const char*   name,
                                            const char*   string1,
                                            const char*   string2,
                                            int           int1,
                                            int           int2)
{
	char* pixmaps_colon_sep = g_strdup(string1);
	char** pixmaps = g_strsplit(pixmaps_colon_sep, ":", 100);
	GtkPixButton* pix_button;
	pix_button = GTK_PIX_BUTTON(gtk_type_new(GTK_TYPE_PIX_BUTTON));
	gtk_pix_button_set_pixmaps_by_filename(pix_button,
	                                       pixmaps[0],
	                                       pixmaps[1],
	                                       pixmaps[2],
	                                       pixmaps[3]);
	g_strfreev(pixmaps);
	g_free(pixmaps_colon_sep);
	return GTK_WIDGET(pix_button);
}

